# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Classes and functions for path ensembles.

The classes and functions defined in this module are useful for
representing path ensembles.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

PathEnsemble (:py:class:`.PathEnsemble`)
    Class for defining path ensembles.

PathEnsembleExt (:py:class:`.PathEnsembleExt`)
    Class for defining path ensembles when we are working with
    paths stored on disk and not in memory only.
"""
import collections
import logging
import os
import shutil
from pyretis.core.path import Path
from pyretis.core.common import big_fat_comparer
from pyretis.core.random_gen import create_random_generator


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['PathEnsemble', 'PathEnsembleExt', 'get_path_ensemble_class',
           'generate_ensemble_name']


def generate_ensemble_name(ensemble_number, zero_pad=3):
    """Generate a simple name for an ensemble.

    The simple name will have a format like 01, 001, 0001 etc. and it
    is used to name the path ensemble and the output directory.

    Parameters
    ----------
    ensemble_number : int
        The number representing the ensemble.
    zero_pad : int, optional
        The number of zeros to use for padding the name.

    Returns
    -------
    out : string
        The ensemble name.

    """
    if zero_pad < 3:
        logger.warning('zero_pad must be >= 3. Setting it to 3.')
        zero_pad = 3
    fmt = f'{{:0{zero_pad}d}}'
    return fmt.format(ensemble_number)


def _generate_file_names(path, target_dir, prefix=None):
    """Generate new file names for moving copying paths.

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        This is the path object we are going to store.
    target_dir : string
        The location where we are moving the path to.
    prefix : string, optional
        The prefix can be used to prefix the name of the files.

    Returns
    -------
    out[0] : list
        A list with new file names.
    out[1] : dict
        A dict which defines the unique "source -> destination" for
        copy/move operations.

    """
    source = {}
    new_pos = []
    for phasepoint in path.phasepoints:
        pos_file, idx = phasepoint.particles.get_pos()
        if pos_file not in source:
            localfile = os.path.basename(pos_file)
            if prefix is not None:
                localfile = f'{prefix}{localfile}'
            dest = os.path.join(target_dir, localfile)
            source[pos_file] = dest
        dest = source[pos_file]
        new_pos.append((dest, idx))
    return new_pos, source


class PathEnsemble:
    """Representation of a path ensemble.

    This class represents a collection of paths in a path ensemble.
    In general, paths may be "long and complicated" so here, we really
    just store a simplified abstraction of the path, which is obtained
    by the `Path.get_path_data()` function for a given `Path` object.
    The returned dictionary is stored in the list `PathEnsemble.paths`.
    The only full path we store is the last accepted path. This is
    convenient for the RETIS method where paths may be swapped between
    path ensembles.

    Attributes
    ----------
    ensemble_number : integer
        This integer is used to represent the path ensemble, for RETIS
        simulations it's useful to identify the path ensemble. The path
        ensembles are numbered sequentially 0, 1, 2, etc. This
        corresponds to ``[0^-]``, ``[0^+]``, ``[1^+]``, etc.
    ensemble_name : string
        A string which can be used for printing the ensemble name.
        This is of form ``[0^-]``, ``[0^+]``, ``[1^+]``, etc.
    ensemble_name_simple : string
        A string with a simpler representation of the ensemble name,
        can be used for creating output files etc.
    interfaces : list of floats
        Interfaces, specified with the values for the
        order parameters: `[left, middle, right]`.
    paths : list
        This list contains the stored information for the paths. Here
        we only store the data returned by calling the `get_path_data()`
        function of the `Path` object.
    nstats : dict of ints
        This dict store some statistics for the path ensemble. The keys
        are:

        * npath : The number of paths stored.
        * nshoot : The number of accepted paths generated by shooting.
        * ACC, BWI, ... : Number of paths with given status (from `_STATUS`).
    maxpath : int
        The maximum number of paths to store.
    last_path : object like :py:class:`.PathBase`
        This is the last **accepted** path.

    """

    def __init__(self, ensemble_number, interfaces,
                 rgen=None, maxpath=10000, exe_dir=None):
        """Initialise the PathEnsemble object.

        Parameters
        ----------
        ensemble_number : integer
            An integer used to identify the ensemble.
        interfaces : list of floats
            These are the interfaces specified with the values
            for the order parameters: ``[left, middle, right]``.
        rgen : object like :py:class:`.RandomGenerator`, optional
            The random generator that will be used for the
            paths that required random numbers.
        maxpath : integer, optional
            The maximum number of paths to store information for in memory.
            Note, that this will not influence the analysis as long as
            you are using the output files when running the analysis.
        exe_dir : string, optional
            The base folder where the simulation was executed from.
            This is used to set up output directories for the path
            ensemble.

        """
        if rgen is None:
            rgen = create_random_generator()
        self.rgen = rgen
        self.ensemble_number = ensemble_number
        self.interfaces = tuple(interfaces)  # Should not change interfaces.
        self.last_path = None
        self.nstats = {'npath': 0, 'nshoot': 0, 'ACC': 0}
        self.paths = []
        self.maxpath = maxpath
        if self.ensemble_number == 0:
            self.ensemble_name = '[0^-]'
            self.start_condition = 'R'
        else:
            ensemble_number = self.ensemble_number - 1
            self.ensemble_name = f'[{ensemble_number}^+]'
            self.start_condition = 'L'
        self.ensemble_name_simple = generate_ensemble_name(
            self.ensemble_number
        )
        self.directory = collections.OrderedDict()
        self.directory['path_ensemble'] = None
        self.directory['accepted'] = None
        self.directory['generate'] = None
        self.directory['traj'] = None
        if exe_dir is not None:
            path_dir = os.path.join(exe_dir, self.ensemble_name_simple)
            self.update_directories(path_dir)

    def __eq__(self, other):
        """Check if two path_ensemble are equal."""
        equal = True
        if self.__class__ != other.__class__:
            logger.debug('%s and %s.__class__ differ', self, other)
            return False

        if set(self.__dict__) != set(other.__dict__):
            logger.debug('%s and %s.__dict__ differ', self, other)
            equal = False

        for i in ['directory', 'interfaces', 'nstats', 'paths']:
            if hasattr(self, i):
                for j, k in zip(getattr(self, i), getattr(other, i)):
                    if j != k:
                        logger.debug('%s for %s and %s attributes are %s and '
                                     '%s', i, self, other, j, k)
                        equal = False

        for i in ['ensemble_name',
                  'ensemble_name_simple', 'ensemble_number',
                  'maxpath', 'start_condition']:
            if hasattr(self, i):
                if getattr(self, i) != getattr(other, i):
                    logger.debug('%s for %s and %s, attributes are %s and %s',
                                 i, self, other,
                                 getattr(self, i), getattr(other, i))
                    equal = False

        if hasattr(self, 'last_path'):
            if self.last_path != other.last_path:
                logger.debug('last paths differs')
                equal = False

        if hasattr(self, 'rgen'):
            if self.rgen.__class__ != other.rgen.__class__:
                logger.debug('self.rgen.__class__ differs')
                return False
            if self.rgen.__dict__['seed'] != other.rgen.__dict__['seed']:
                logger.debug('rgen seed differs')
                equal = False
            if not big_fat_comparer(self.rgen.__dict__['rgen'].get_state(),
                                    other.rgen.__dict__['rgen'].get_state()):
                logger.debug('rgen differs')
                equal = False

        return equal

    def __ne__(self, other):
        """Check if two paths are not equal."""
        return not self == other

    def directories(self):
        """Yield the directories PyRETIS should make."""
        for key in self.directory:
            yield self.directory[key]

    def update_directories(self, path):
        """Update directory names.

        This method will not create new directories, but it will
        update the directory names.

        Parameters
        ----------
        path : string
            The base path to set.

        """
        for key, val in self.directory.items():
            if key == 'path_ensemble':
                self.directory[key] = path
            else:
                self.directory[key] = os.path.join(path, key)
            if val is None:
                logger.debug('Setting directory "%s" to %s', key,
                             self.directory[key])
            else:
                logger.debug('Updating directory "%s": %s -> %s',
                             key, val, self.directory[key])

    def reset_data(self):
        """Erase the stored data in the path ensemble.

        It can be used in combination with flushing the data to a
        file in order to periodically write and empty the amount of data
        stored in memory.

        Notes
        -----
        We do not reset `self.last_path` as this might be used in the
        RETIS function.

        """
        self.paths = []
        for key in self.nstats:
            self.nstats[key] = 0

    def store_path(self, path):
        """Store a new accepted path in the path ensemble.

        Parameters
        ----------
        path : object like :py:class:`.PathBase`
            The path we are going to store.

        Returns
        -------
        None, but we update `self.last_path`.

        """
        self.last_path = path

    def add_path_data(self, path, status, cycle=0):
        """Append data from the given path to `self.path_data`.

        This will add the data from a given` path` to the list path data
        for this ensemble. If will also update `self.last_path` if the
        given `path` is accepted.

        Parameters
        ----------
        path : object like :py:class:`.PathBase`
            This is the object to store data from.
        status : string
            This is the status of the path. Note that the path object
            also has a status property. However, this one might not be
            set, for instance when the path is just None. We therefore
            use `status` here as a parameter.
        cycle : int, optional
            The current cycle number.

        """
        if len(self.paths) >= self.maxpath:
            # This is just to limit the data we keep in memory in
            # case of really long simulations.
            logger.debug(('Path-data memory storage reset for ensemble %s.\n'
                          'This is just to limit the amount of data we store '
                          'in memory.\nThis will *NOT* influence the '
                          'simulation'), self.ensemble_name)
            self.paths = []
        # Update statistics:
        if path is None:
            # Here we add a dummy path with minimal info. This is because we
            # could not generate a path for some reason which should be
            # specified by the status.
            path_data = {'status': status, 'generated': ('', 0, 0, 0),
                         'weight': 1.}
        else:
            path_data = path.get_path_data(status, self.interfaces)
            if 'EXP' in status:
                path_data['status'] = 'EXP'
            if path_data['status'] in {'ACC', 'EXP'}:  # Store the path:
                self.store_path(path)
                if path_data['generated'][0] in {'sh', 'ss', 'wt', 'wf'}:
                    self.nstats['nshoot'] += 1
        path_data['cycle'] = cycle  # Also store cycle number.
        self.paths.append(path_data)  # Store the new data.
        # Update some statistics:
        # This is to count also for the first occurrence of the status:
        self.nstats[status] = self.nstats.get(status, 0) + 1
        self.nstats['npath'] += 1

    def get_accepted(self):
        """Yield accepted paths from the path ensemble.

        This function will return an iterator useful for iterating over
        accepted paths only. In the path ensemble we store both accepted
        and rejected paths. This function will loop over all paths
        stored and yield the accepted paths the correct number of times.
        """
        last_path = None
        for path in self.paths:
            if path['status'] == 'ACC':
                last_path = path
            yield last_path

    def get_acceptance_rate(self):
        """Return acceptance rate for the path ensemble.

        The acceptance rate is obtained as the fraction of accepted
        paths to the total number of paths in the path ensemble. This
        will only consider the paths that are currently stored in
        `self.paths`.

        Returns
        -------
        out : float
            The acceptance rate.

        """
        acc = 0
        npath = 0
        for path in self.paths:
            if path['status'] == 'ACC':
                acc += 1
            npath += 1
        return float(acc) / float(npath)

    def get_paths(self):
        """Yield the different paths stored in the path ensemble.

        It is included here in order to have a simple compatibility
        between the :py:class:`.PathEnsemble` object and the
        py:class:`.PathEnsembleFile` object. This is useful for the
        analysis.

        Yields
        ------
        out : dict
            This is the dictionary representing the path data.

        """
        for path in self.paths:
            yield path

    def move_path_to_generate(self, _path, _prefix=None):
        """Move a path for temporary storing."""
        return

    def copy_path_to_generate(self, path, _prefix=None):
        """Copy a path for temporary storing.

        Parameters
        ----------
        path : object like :py:class:`.PathBase`
            The path to copy.

        Returns
        -------
        out : object like :py:class:`.PathBase`
            The copy of the path.

        """
        path_copy = path.copy()
        return path_copy

    def __str__(self):
        """Return a string with some info about the path ensemble."""
        msg = [f'Path ensemble: {self.ensemble_name}']
        msg += [f'\tInterfaces: {self.interfaces}']
        if self.nstats['npath'] > 0:
            npath = self.nstats['npath']
            nacc = self.nstats.get('ACC', 0)
            msg += [f'\tNumber of paths stored: {npath}']
            msg += [f'\tNumber of accepted paths: {nacc}']
            ratio = float(nacc) / float(npath)
            msg += [f'\tRatio accepted/total paths: {ratio}']
        return '\n'.join(msg)

    def restart_info(self):
        """Return a dictionary with restart information."""
        restart = {
            'nstats': self.nstats,
            'interfaces': self.interfaces,
            'ensemble_number': self.ensemble_number,
        }
        if hasattr(self, 'rgen'):
            restart['rgen'] = self.rgen.get_state()

        if self.last_path:
            restart['last_path'] = self.last_path.restart_info()

        return restart

    def load_restart_info(self, info, cycle=0):
        """Load restart information.

        Parameters
        ----------
        info : dict
            A dictionary with the restart information.
        cycle : integer, optional
            The current simulation cycle.

        """
        self.nstats = info['nstats']
        for attr in ('interfaces', 'ensemble_number'):
            if info[attr] != getattr(self, attr):
                logger.warning(
                    'Inconsistent path ensemble restart info for %s', attr)
        for key in info:
            if key == 'rgen':
                self.rgen = create_random_generator(info[key])
            elif key == 'last_path':
                rgen = create_random_generator(info[key]['rgen'])
                path = Path(rgen=rgen)
                path.load_restart_info(info['last_path'])
                path_data = path.get_path_data('ACC', self.interfaces)
                path_data['cycle'] = cycle
                self.last_path = path
                self.paths.append(path_data)
            elif hasattr(self, key):
                setattr(self, key, info[key])

    def clear_generate(self):
        """Remove all the files in an ensemble/generate/ directory.

        This is toggled on by adding 'remove_generate = True' to the retis.rst
        input file, under the [simulation] section.

        """
        gendir = self.directory['generate']
        logger.debug("Removing generate files from %s", gendir)
        if gendir is not None and os.path.exists(gendir):
            for file in os.listdir(gendir):
                path = os.path.join(gendir, file)
                logger.debug("Removing generate file %s", path)
                # assert that path ends with generate/file
                assert path.endswith(os.path.join('generate', file))
                os.remove(path)


class PathEnsembleExt(PathEnsemble):
    """Representation of a path ensemble.

    This class is similar to :py:class:`.PathEnsemble` but it is made
    to work with external paths. That is, some extra file handling is
    done when accepting a path.

    """

    @staticmethod
    def _move_path(path, target_dir, prefix=None):
        """Move a path to a given target directory.

        Parameters
        ----------
        path : object like :py:class:`.PathBase`
            This is the path object we are going to move.
        target_dir : string
            The location where we are moving the path to.
        prefix : string, optional
            To give a prefix to the name of moved files.

        """
        logger.debug('Moving path to %s', target_dir)
        new_pos, source = _generate_file_names(path, target_dir,
                                               prefix=prefix)
        for pos, phasepoint in zip(new_pos, path.phasepoints):
            phasepoint.particles.set_pos(pos)
        for src, dest in source.items():
            if src == dest:
                logger.debug('Skipping move %s -> %s', src, dest)
            else:
                if os.path.exists(dest):
                    if os.path.isfile(dest):
                        logger.debug('Removing %s as it exists', dest)
                        os.remove(dest)
                logger.debug('Moving %s -> %s', src, dest)
                os.rename(src, dest)

    @staticmethod
    def _copy_path(path, target_dir, prefix=None):
        """Copy a path to a given target directory.

        Parameters
        ----------
        path : object like :py:class:`.PathBase`
            This is the path object we are going to copy.
        target_dir : string
            The location where we are copying the path to.
        prefix : string, optional
            To give a prefix to the name of copied files.

        Returns
        -------
        out : object like py:class:`.PathBase`
            A copy of the input path.

        """
        path_copy = path.copy()
        new_pos, source = _generate_file_names(path_copy, target_dir,
                                               prefix=prefix)
        # Update positions:
        for pos, phasepoint in zip(new_pos, path_copy.phasepoints):
            phasepoint.particles.set_pos(pos)
        for src, dest in source.items():
            if src != dest:
                if os.path.exists(dest):
                    if os.path.isfile(dest):
                        logger.debug('Removing %s as it exists', dest)
                        os.remove(dest)
                logger.debug('Copy %s -> %s', src, dest)
                shutil.copy(src, dest)
        return path_copy

    def store_path(self, path):
        """Store a path by explicitly moving it.

        Parameters
        ----------
        path : object like :py:class:`.PathBase`
            This is the path object we are going to store.

        """
        self._move_path(path, self.directory['accepted'])
        self.last_path = path
        for entry in self.list_superfluous():
            try:
                os.remove(entry)
            except OSError:  # pragma: no cover
                pass

    def list_superfluous(self):
        """List files in accepted directory that we do not need."""
        last = set()
        if self.last_path:
            for phasepoint in self.last_path.phasepoints:
                pos_file, _ = phasepoint.particles.get_pos()
                last.add(pos_file)
        for entry in os.scandir(self.directory['accepted']):
            if entry.is_file() and entry.path not in last:
                yield entry.path

    def move_path_to_generate(self, path, prefix=None):
        """Move a path for temporary storing."""
        self._move_path(path, self.directory['generate'], prefix=prefix)

    def copy_path_to_generate(self, path, pref=None):
        """Copy a path for temporary storing."""
        return self._copy_path(path, self.directory['generate'], prefix=pref)

    def load_restart_info(self, info, cycle=0):
        """Load restart for external path."""
        super().load_restart_info(info, cycle=cycle)
        # Update file names:
        directory = self.directory['accepted']
        for phasepoint in self.last_path.phasepoints:
            filename = os.path.basename(phasepoint.particles.get_pos()[0])
            new_file_name = os.path.join(directory, filename)
            if not os.path.isfile(new_file_name):
                logger.critical('The restart path "%s" does not exist',
                                new_file_name)
            phasepoint.particles.set_pos((new_file_name,
                                          phasepoint.particles.get_pos()[1]))


def get_path_ensemble_class(ensemble_type):
    """Return the path ensemble class consistent with the given engine.

    Parameters
    ----------
    ensemble_type : string
        The type of ensemble we are requesting.

    """
    path_ensemble_map = {'internal': PathEnsemble,
                         'external': PathEnsembleExt}
    try:
        return path_ensemble_map[ensemble_type]
    except KeyError as err:
        msg = f'Unknown ensemble type "{ensemble_type}" requested.'
        logger.critical(msg)
        raise ValueError(msg) from err



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test parsing from a settings input file.

Here we test that we parse the input file correctly and also that
we fail in predictable ways.
"""
import logging
import numpy as np
import os
import tempfile
import unittest
from io import StringIO
from pyretis.core.common import big_fat_comparer
from pyretis.core.units import create_conversion_factors, CONVERT
from pyretis.engines import Verlet, VelocityVerlet, Langevin
from pyretis.forcefield import PotentialFunction
from pyretis.forcefield.potentials import (
    DoubleWell,
    DoubleWellWCA,
    PairLennardJonesCut,
    PairLennardJonesCutnp,
    RectangularWell
)
from pyretis.inout.settings import (
    copy_settings,
    look_for_input_files,
    parse_settings_file,
    settings_to_text,
    write_settings_file,
    add_default_settings,
    add_specific_default_settings,
    _clean_settings,
    _parse_raw_section,
    _parse_all_raw_sections,
    _parse_sections,
    SECTIONS
)
from pyretis.orderparameter import (
    Angle,
    Dihedral,
    Distance,
    Distancevel,
    DistanceVelocity,
    OrderParameter,
    Position,
    PositionVelocity,
    Velocity,
)
from pyretis.setup.common import (
    create_engine,
    create_orderparameter
)
from pyretis.setup.createforcefield import (
    create_potentials,
    create_force_field
)
from pyretis.setup.createsystem import create_initial_positions
from unittest.mock import patch
from .help import turn_on_logging

logging.disable(logging.CRITICAL)

LOCAL_DIR = os.path.abspath(os.path.dirname(__file__))


def _read_raw_settings(filename):
    """Read raw settings from a local file.

    Parameters
    ----------
    filename : string
        The file we are going to read.

    Return
    ------
    out : string
        The data read from the file.
    """
    inputfile = os.path.join(LOCAL_DIR, filename)
    data = None
    with open(inputfile, 'r', encoding='utf-8') as indata:
        data = indata.read()
    return data


def _test_correct_parsing(test, data, correct):
    """Helper method to test that we correctly parse settings.

    Parameters
    ----------
    test : object like unittest.TestCase
        The test that should pass or fail.
    data : string
        Raw data to be parsed.
    correct : dict
        The correct data.

    Returns
    -------
    out : dict
        The parsed settings.
    """
    raw = _parse_sections(data.split('\n'))
    settings = _parse_all_raw_sections(raw)
    for key in settings:
        test.assertEqual(settings[key], correct[key])
    return settings


class KeywordParsing(unittest.TestCase):
    """Test the parsing of input settings."""

    def test_parse_file(self):
        """Test that we can parse an input file."""
        inputfile = os.path.join(LOCAL_DIR, 'settings.rst')
        settings = parse_settings_file(inputfile)
        correct = {
            'heading': {'text': ('Molecular dynamics example settings\n'
                                 '===================================\n'
                                 'this is a simple test system to check '
                                 'the input statements are correct')},
            'system': {'units': 'lj',
                       'dimensions': 3,  # Added by default.
                       'temperature': 2.0},
            'simulation': {'task': 'md-nve',
                           'steps': 100,
                           'rgen': 'rgen',
                           'priority_shooting': False},
            'engine': {'class': 'velocityverlet',
                       'timestep': 0.002,
                       'rgen': 'rgen',
                       'type': 'internal'},
            'particles': {'position': {'input_file': 'initial.gro'},
                          'velocity': {'generate': 'maxwell',
                                       'temperature': 2.0,
                                       'momentum': True,
                                       'seed': 0},
                          'mass': {'Ar': 1.0},
                          'type': 'internal'},
            'forcefield': {'description': 'Lennard Jones test'},
            'potential': [{'class': 'PairLennardJonesCutnp',
                           'shift': True}],
        }
        del settings['engine']['exe_path']
        del settings['simulation']['exe_path']

        for key in correct:
            self.assertIn(key, settings)
            self.assertEqual(correct[key], settings[key])

    def test_keyword_format(self):
        """Test different forms of some simple keywords."""
        test_data = [
            (["units = 'lj'"], {'units': 'lj'}),  # Normal input
            (["unITS = 'lj'"], {'units': 'lj'}),  # Test case-sensitive?
            (["units = 'lj' # comment"], {'units': 'lj'}),  # Add a comment.
            (["units = 'lj' # comment units='a'"], {'units': 'lj'}),
            (["unITS='lj'"], {'units': 'lj'}),  # Test spacing.
            (["unITS= 'lj'"], {'units': 'lj'}),  # Test spacing.
            (['units = "lj"'], {'units': 'lj'}),  # Test " vs '.
            (['units =          "lj"'], {'units': 'lj'}),  # Test spacing.
            (['units        = "lj"'], {'units': 'lj'}),  # Test spacing.
            (['units = lj'], {'units': 'lj'}),  # Test quotations.
            (['units = "lj'], {'units': '"lj'})  # Test quotations.
        ]

        for data in test_data:
            setting = _parse_raw_section(data[0], 'system')
            self.assertEqual(setting, data[1])

    def test_keyword_dict(self):
        """Test some cases when reading dicts."""
        teststr = []
        correct = []
        # First, some simple input:
        teststr.append(['Engine settings',
                        '---------------',
                        'class = velocityverlet', 'timestep = 0.002'])
        correct.append({'timestep': 0.002, 'class': 'velocityverlet'})
        # Test input with comment(s):
        teststr.append(['Engine settings',
                        '---------------',
                        'class = velocityverlet', 'timestep = 0.002 # fs'])
        correct.append({'timestep': 0.002, 'class': 'velocityverlet'})
        # Test with white space, etc.:
        teststr.append(['Engine settings',
                        '---------------', 'junk',
                        '    ', 'junk = 10',
                        'class =    velocityverlet', 'timestep=0.002'])
        correct.append({'timestep': 0.002, 'junk': 10,
                        'class': 'velocityverlet'})
        teststr.append(['Engine settings',
                        '---------------', 'junk',
                        '    ', 'junk = 10',
                        'class =    velocityverlet', 'timestep=0.002'])
        correct.append({'timestep': 0.002, 'class': 'velocityverlet',
                        'junk': 10})
        teststr.append(['Engine settings',
                        '---------------',
                        'class = langevin', 'timestep = 0.002',
                        'gamma = 0.3', 'high_friction = False',
                        'seed = 0'])
        correct.append({'timestep': 0.002, 'class': 'langevin',
                        'gamma': 0.3, 'high_friction': False,
                        'seed': 0})

        for tst, corr in zip(teststr, correct):
            raw = _parse_sections(tst)
            setting = _parse_raw_section(raw['engine'], 'engine')
            self.assertEqual(setting, corr)

    def test_write_and_read(self):
        """Test that we can parse some data, write it and read it."""
        data = _read_raw_settings('settings-rw-test.rst')
        correct = {'engine': {'timestep': 0.002,
                              'class': 'velocityverlet'},
                   'system': {'dimensions': 2, 'temperature': 1.0},
                   'simulation': {'task': 'md-nve', 'steps': 100}}
        raw = _parse_sections(data.split('\n'))
        settings = _parse_all_raw_sections(raw)
        settings = _test_correct_parsing(self, data, correct)
        with tempfile.NamedTemporaryFile() as temp:
            txt = settings_to_text(settings)
            temp.write(txt.encode('utf-8'))
            temp.flush()
            settings_read = parse_settings_file(temp.name, add_default=False)
        self.assertEqual(settings_read, correct)

    def test_ignore_section(self):
        """Test that we ignore unknown sections."""
        data = _read_raw_settings('settings-unknown-section.rst')
        correct = {'engine': {'timestep': 0.002,
                              'class': 'velocityverlet'}}
        settings = _test_correct_parsing(self, data, correct)
        self.assertNotIn('junk', settings)


class KeywordEngine(unittest.TestCase):
    """Test the parsing of input settings for engines."""

    def test_load_external_engine(self):
        """Test that we can load external python modules for engines."""
        data = _read_raw_settings('external-engine.rst')
        correct = {'engine': {'class': 'FooEngine',
                              'module': 'fooengine.py',
                              'timestep': 0.5,
                              'extra': 100}}
        settings = _test_correct_parsing(self, data, correct)
        # Here we add the exe_path key to the settings to tell
        # PyRETIS where we are executing from. This is to locate the
        # script we want to run.
        settings['simulation'] = {'exe_path': LOCAL_DIR}
        fooengine = create_engine(settings)
        self.assertEqual(fooengine.timestep,
                         correct['engine']['timestep'])
        self.assertEqual(fooengine.extra,  # pylint: disable=no-member
                         correct['engine']['extra'])

    def test_fail_external_engine(self):
        """Test that external loads fail in a predicable way."""
        # First test: an engine that forgot to define a required method.
        test_data, correct = [], []
        test_data.append('---------------\n'
                         'Engine settings\n'
                         '---------------\n'
                         'class = BarEngine\n'
                         'module = fooengine.py')
        correct.append({'engine': {'class': 'BarEngine',
                                   'module': 'fooengine.py'}})
        test_data.append('Engine settings\n'
                         '---------------\n'
                         'class = BazEngine\n'
                         'module = fooengine.py')
        correct.append({'engine': {'class': 'BazEngine',
                                   'module': 'fooengine.py'}})
        test_data.append('Engine\n'
                         '------\n'
                         'module =  dummy')
        correct.append({'engine': {'module': 'dummy'}})
        test_data.append('Engine\n'
                         '------\n'
                         'module = dummy\n'
                         'class = dummy')
        correct.append({'engine': {'module': 'dummy', 'class': 'dummy'}})

        for data, corr in zip(test_data, correct):
            settings = _test_correct_parsing(self, data, corr)
            settings['simulation'] = {'exe_path': LOCAL_DIR}
            with self.assertRaises(ValueError):
                create_engine(settings)

    def test_internal_engine(self):
        """Test that we can load all internal engines."""
        klass, test_data, correct = [], [], []
        klass.append(Verlet)
        test_data.append('Engine\n'
                         '------\n'
                         'class = Verlet\n'
                         'timestep = 0.5')
        correct.append({'engine': {'class': 'Verlet', 'timestep': 0.5}})
        klass.append(VelocityVerlet)
        test_data.append('Engine\n'
                         '------\n'
                         'class = VelocityVerlet\n'
                         'timestep = 0.314\n'
                         'desc = Test VV engine')
        correct.append({'engine': {'class': 'VelocityVerlet',
                                   'timestep': 0.314,
                                   'desc': 'Test VV engine'}})
        klass.append(Langevin)
        test_data.append('Engine\n'
                         '------\n'
                         'class = Langevin\n'
                         'timestep = 0.1\n'
                         'gamma = 2.718281828\n'
                         'seed = 101\n'
                         'high_friction = True')
        correct.append({'engine': {'class': 'Langevin',
                                   'timestep': 0.1,
                                   'gamma': 2.718281828,
                                   'seed': 101,
                                   'high_friction': True}})
        klass.append(Langevin)
        test_data.append('Engine\n'
                         '------\n'
                         'class = Langevin\n'
                         'timestep = 0.25\n'
                         'gamma = 2.718281828\n'
                         'seed = 11\n'
                         'high_friction = False')
        correct.append({'engine': {'class': 'Langevin',
                                   'timestep': 0.25,
                                   'gamma': 2.718281828,
                                   'seed': 11,
                                   'high_friction': False}})
        for data, corr, cls in zip(test_data, correct, klass):
            settings = _test_correct_parsing(self, data, corr)
            engine = create_engine(settings)
            self.assertIsInstance(engine, cls)
            self.assertAlmostEqual(engine.timestep,
                                   corr['engine']['timestep'])
            for key, val in corr['engine'].items():
                if hasattr(engine, key):
                    self.assertAlmostEqual(getattr(engine, key), val)


class KeywordOrderPrameter(unittest.TestCase):
    """Test creation of order parameters."""

    def test_load_orderparameter(self):
        """Test loading of external order parameter."""
        data = [
            'Orderparameter',
            '--------------',
            'class = FooOrderParameter',
            'module = fooorderparameter.py',
            'name = Dummy'
        ]
        correct = {
            'orderparameter': {
                'class': 'FooOrderParameter',
                'module': 'fooorderparameter.py',
                'name': 'Dummy'
            }
        }
        settings = _test_correct_parsing(self, '\n'.join(data), correct)
        # Here we add the exe-path key to the settings to tell
        # PyRETIS where we are executing from. This is to locate the
        # script we want to run.
        settings['simulation'] = {'exe_path': LOCAL_DIR}
        orderp = create_orderparameter(settings)
        self.assertEqual(
            correct['orderparameter']['class'],
            orderp.__class__.__name__,
        )

    def test_fail_orderparameter(self):
        """Test that loading external order parameters fails."""
        test_data, correct = [], []
        test_data.append('Orderparameter\n'
                         '--------------\n'
                         'class = BarOrderParameter\n'
                         'module = fooorderparameter.py')
        correct.append(
            {'orderparameter': {'class': 'BarOrderParameter',
                                'module': 'fooorderparameter.py'}}
        )
        for data, corr in zip(test_data, correct):
            settings = _test_correct_parsing(self, data, corr)
            settings['simulation'] = {'exe_path': LOCAL_DIR}
            with self.assertRaises(ValueError):
                create_orderparameter(settings)

    def test_create_orderparameter(self):
        """Test that we can create internal order parameters."""
        test_data = _read_raw_settings('internal-order.rst')
        klass = [
            OrderParameter,
            Position,
            Velocity,
            PositionVelocity,
            Distance,
            Distancevel,
            DistanceVelocity,
            Angle,
            Dihedral,
        ]
        correct = {
            'orderparameter': {
                'class': 'OrderParameter',
                'name': 'test'
            },
            'collective-variable':
                [
                    {
                        'class': 'Position',
                        'name': 'Position',
                        'index': 0, 'dim': 'x',
                        'periodic': False
                    },
                    {
                        'class': 'Velocity',
                        'name': 'Velocity',
                        'index': 0, 'dim': 'x',
                    },
                    {
                        'class': 'PositionVelocity',
                        'name': 'PositionVelocity',
                        'index': 0, 'dim': 'x',
                        'periodic': True
                    },
                    {
                        'class': 'Distance',
                        'name': 'My distance',
                        'index': (100, 101), 'periodic': False
                    },
                    {
                        'class': 'Distancevel',
                        'name': 'My distance Velocity',
                        'index': (102, 103), 'periodic': True
                    },
                    {
                        'class': 'DistanceVelocity',
                        'name': 'My distance Position Velocity',
                        'index': (1001, 1002), 'periodic': True
                    },
                    {
                        'class': 'Angle',
                        'name': 'Angle',
                        'index': (1001, 1002, 1003), 'periodic': True
                    },
                    {
                        'class': 'Dihedral',
                        'name': 'Dihedral',
                        'index': (1001, 1002, 1003, 1004), 'periodic': True
                    },
                ],
        }
        raw = _parse_sections(test_data.split('\n'))
        settings = _parse_all_raw_sections(raw)
        # Compare for order parameter:
        for key, val in settings['orderparameter'].items():
            self.assertEqual(val, correct['orderparameter'][key])
        for setting, corr in zip(settings['collective-variable'],
                                 correct['collective-variable']):
            for key, val in setting.items():
                self.assertEqual(val, corr[key])
        order = create_orderparameter(settings)
        for i, j in zip(order.order_parameters, klass):
            self.assertIsInstance(i, j)


class KeywordParticles(unittest.TestCase):
    """Test initialisation of particles."""

    def test_lattice(self):
        """Test initialisation on a lattice."""
        data = _read_raw_settings('settings-lattice.rst')
        correct = {'particles': {'position': {'generate': 'fcc',
                                              'repeat': [6, 6, 6],
                                              'lcon': 1.0}},
                   'system': {'units': 'lj'}}
        settings = _test_correct_parsing(self, data, correct)
        create_conversion_factors(settings['system']['units'])
        particles, size, _ = create_initial_positions(settings)
        correct_size = {'low': np.array([0., 0., 0.]),
                        'high': np.array([6., 6., 6.])}
        for key, val in correct_size.items():
            self.assertTrue(np.allclose(size[key], val))
        self.assertEqual(particles.npart, 4 * 6 * 6 * 6)
        self.assertAlmostEqual(particles.mass[0][0], 1.0)
        self.assertAlmostEqual(particles.imass[0][0], 1.0)
        for i in range(particles.npart):
            self.assertEqual(particles.name[i], 'Ar')

    def test_lattice_type(self):
        """Test initialisation on a lattice with types."""
        data = _read_raw_settings('settings-lattice-type.rst')
        correct = {'particles': {'position': {'generate': 'fcc',
                                              'repeat': [3, 3, 3],
                                              'lcon': 1.0},
                                 'ptype': [0, 1]},
                   'system': {'units': 'lj'}}
        settings = _test_correct_parsing(self, data, correct)
        particles, _, _ = create_initial_positions(settings)
        for i in range(particles.npart):
            self.assertEqual(particles.name[i], 'Ar')
        for i in range(particles.npart):
            if i == 0:
                self.assertEqual(particles.ptype[i], 0)
            else:
                self.assertEqual(particles.ptype[i], 1)

    def test_lattice_dens(self):
        """Test initialisation on a lattice with density set."""
        data = _read_raw_settings('settings-lattice-dens.rst')
        correct = {'particles': {'position': {'generate': 'fcc',
                                              'repeat': [3, 3, 3],
                                              'density': 0.9}},
                   'system': {'units': 'lj'}}
        settings = _test_correct_parsing(self, data, correct)
        particles, size, _ = create_initial_positions(settings)
        correct_size = []
        lcon = 3.0 * (4.0 / 0.9)**(1.0 / 3.0)
        for _ in settings['particles']['position']['repeat']:
            correct_size.append([0.0, lcon])
        correct_size = np.array(correct_size)
        self.assertTrue(np.allclose(size['low'], correct_size[:, 0]))
        self.assertTrue(np.allclose(size['high'], correct_size[:, 1]))
        for i in range(particles.npart):
            self.assertEqual(particles.name[i], 'Ar')
            self.assertEqual(particles.ptype[i], 0)

    def test_lattice_dens_lcon(self):
        """Test initialisation on a lattice with density and lcon set."""
        data = _read_raw_settings('settings-lattice-dens-lcon.rst')
        correct = {'particles': {'position': {'generate': 'fcc',
                                              'repeat': [3, 3, 3],
                                              'density': 0.9,
                                              'lcon': 1000.}},
                   'system': {'units': 'lj'}}
        settings = _test_correct_parsing(self, data, correct)
        _, size, _ = create_initial_positions(settings)
        correct_size = []
        # `lcon` should be replaced by density:
        lcon = 3.0 * (4.0 / 0.9)**(1.0 / 3.0)
        for _ in settings['particles']['position']['repeat']:
            correct_size.append([0.0, lcon])
        correct_size = np.array(correct_size)
        self.assertTrue(np.allclose(size['low'], correct_size[:, 0]))
        self.assertTrue(np.allclose(size['high'], correct_size[:, 1]))

    def test_lattice_and_mass(self):
        """Test initialisation on a lattice and setting of masses/types."""
        data = _read_raw_settings('settings-lattice-mass.rst')
        correct = {'particles': {'position': {'generate': 'fcc',
                                              'repeat': [3, 3, 3],
                                              'lcon': 1.},
                                 'ptype': [0, 1],
                                 'name': ['Ar', 'Kr'],
                                 'mass': {'Ar': 1.0}},
                   'system': {'units': 'lj'}}
        settings = _test_correct_parsing(self, data, correct)
        particles, _, _ = create_initial_positions(settings)
        for i in range(particles.npart):
            if i == 0:
                self.assertEqual(particles.ptype[i], 0)
                self.assertEqual(particles.name[i], 'Ar')
                self.assertAlmostEqual(particles.mass[i][0], 1.0)
            else:
                self.assertEqual(particles.ptype[i], 1)
                self.assertEqual(particles.name[i], 'Kr')
                self.assertAlmostEqual(particles.mass[i][0], 2.09767698)

    def test_inconsistent_dimlattice(self):
        """Test initialisation on a lattice with inconsistent dimensions."""
        data = _read_raw_settings('settings-lattice-dims.rst')
        correct = {'particles': {'position': {'generate': 'sq',
                                              'repeat': [6, 6],
                                              'lcon': 1.},
                                 'name': ['Ar'],
                                 'mass': {'Ar': 1.0}},
                   'system': {'units': 'lj', 'dimensions': 3}}
        settings = _test_correct_parsing(self, data, correct)
        args = [settings]
        self.assertRaises(ValueError, create_initial_positions, *args)

    def test_file_xyz(self):
        """Test initialisation from a XYZ file."""
        data = _read_raw_settings('initial-xyz2.rst')
        correct = {'particles': {'position': {'input_file': 'config.xyz'}},
                   'system': {'units': 'lj'}}
        settings = _test_correct_parsing(self, data, correct)
        units = settings['system']['units']
        create_conversion_factors(units)
        # Add path to the file for this test:
        settings['simulation'] = {'exe_path': LOCAL_DIR}
        particles, size, vel_read = create_initial_positions(settings)
        self.assertFalse(vel_read)
        self.assertIsNone(size)
        pos = particles.pos * CONVERT['length'][units, 'A']
        correct_pos = np.array([[0.0, 0.0, 0.0], [0.5, 0.5, 0.5],
                                [0.5, 0.5, 0.0], [0.5, 0.0, 0.5],
                                [0.0, 0.5, 0.5]])
        self.assertTrue(np.allclose(pos, correct_pos))
        self.assertTrue(all([i == j for i, j in zip(particles.ptype,
                                                    [0, 1, 2, 2, 2])]))
        self.assertTrue(all([i == j for i, j in zip(particles.name,
                                                    ['Ba', 'Hf', 'O',
                                                     'O', 'O'])]))
        masses = []
        for i in particles.mass:
            masses.append(i[0] * CONVERT['mass'][units, 'g/mol'])
        self.assertTrue(np.allclose(masses, [137.327, 178.49, 15.9994,
                                             15.9994, 15.9994]))

    def test_file_gro(self):
        """Test initialisation from a GRO file."""
        data = _read_raw_settings('initial-gro.rst')
        correct = {'particles': {'position': {'input_file': 'config.gro'}},
                   'system': {'units': 'gromacs'}}
        settings = _test_correct_parsing(self, data, correct)
        # Add path to the file for this test:
        create_conversion_factors(settings['system']['units'])
        settings['simulation'] = {'exe_path': LOCAL_DIR}
        particles, size, vel_read = create_initial_positions(settings)
        self.assertTrue(vel_read)
        self.assertTrue(np.allclose(size['cell'], [2., 2., 2.]))
        correct_pos = np.array([[0., 0., 0.], [0.05, 0.05, 0.05],
                                [0.05, 0.05, 0.], [0.05, 0., 0.05],
                                [0., 0.05, 0.05]])
        self.assertTrue(np.allclose(particles.pos, correct_pos))
        self.assertTrue(all([i == j for i, j in zip(particles.ptype,
                                                    [0, 1, 2, 2, 2])]))
        self.assertTrue(all([i == j for i, j in zip(particles.name,
                                                    ['Ba', 'Hf', 'O',
                                                     'O', 'O'])]))
        self.assertTrue(np.allclose(particles.mass.T,
                                    [137.327, 178.49, 15.9994,
                                     15.9994, 15.9994]))
        correct_vel = np.array([[1.0, 1.0, 1.0], [-1.0, -1.0, -1.0],
                                [2.0, 0.0, -2.0], [-2.0, 1.0, 2.0],
                                [0.0, -1.0, 0.0]])
        self.assertTrue(np.allclose(particles.vel, correct_vel))

    def test_file_xyztab(self):
        """Test initialisation from a XYZ file with mass dict."""
        data = _read_raw_settings('initial-xyz.rst')
        correct = {'particles': {'position': {'input_file': 'configtag.xyz'},
                                 'ptype': [0, 0, 0, 1, 1],
                                 'mass': {'Ar': 1., 'Kr': 2.09767698,
                                          'Kr2': 2.09767698}},
                   'system': {'units': 'lj'}}
        settings = _test_correct_parsing(self, data, correct)
        units = settings['system']['units']
        create_conversion_factors(units)
        # Add path to the file for this test:
        settings['simulation'] = {'exe_path': LOCAL_DIR}
        particles, size, vel_read = create_initial_positions(settings)
        self.assertFalse(vel_read)
        self.assertIsNone(size)
        self.assertTrue(all([i == j for i, j in zip(particles.ptype,
                                                    [0, 0, 0, 1, 1])]))
        self.assertTrue(all([i == j for i, j in zip(particles.name,
                                                    ['Ar', 'Ar', 'Ar',
                                                     'Kr', 'Kr2'])]))
        masses = []
        for i in particles.mass:
            masses.append(i[0] * CONVERT['mass'][units, 'g/mol'])
        self.assertTrue(np.allclose(masses, [39.948, 39.948, 39.948,
                                             83.798, 83.798]))


class KeywordForcefield(unittest.TestCase):
    """Test initialisation of force fields."""

    def test_forcefield(self):
        """Test initialisation of a simple force field."""
        data = _read_raw_settings('forcefield-simple.rst')
        correct = {'forcefield': {'description': 'My first force field'},
                   'potential': [{'class': 'PairLennardJonesCutnp',
                                  'shift': True,
                                  'parameter': {0: {'sigma': 1.0,
                                                    'epsilon': 1.0,
                                                    'rcut': 2.5}}}]}
        settings = _test_correct_parsing(self, data, correct)
        forcefield = create_force_field(settings)
        self.assertIsInstance(forcefield.potential[0], PairLennardJonesCutnp)
        self.assertEqual(forcefield.potential[0].shift,
                         correct['potential'][0]['shift'])

    def test_potential_parse(self):
        """Test creation of potentials while parsing input."""
        data = _read_raw_settings('settings-potential1.rst')
        correct = {'potential': [{'class': 'PairLennardJonesCut',
                                  'shift': False, 'mixing': 'geometric',
                                  'parameter': {0: {'sigma': 1.0,
                                                    'epsilon': 1.0,
                                                    'rcut': 2.5},
                                                1: {'sigma': 2.0,
                                                    'epsilon': 2.0,
                                                    'rcut': 2.5}}}]}
        settings = _test_correct_parsing(self, data, correct)
        potentials, pot_par = create_potentials(settings)
        self.assertIsInstance(potentials[0], PairLennardJonesCut)
        self.assertEqual(potentials[0].shift,
                         correct['potential'][0]['shift'])
        # Test that we can assign parameters:
        for pot, params in zip(potentials, pot_par):
            pot.set_parameters(params)
        potparam = potentials[0].params
        self.assertAlmostEqual(potparam[(0, 0)]['epsilon'], 1.0)
        self.assertAlmostEqual(potparam[(0, 0)]['sigma'], 1.0)
        self.assertAlmostEqual(potparam[(0, 0)]['rcut'], 2.5)
        self.assertAlmostEqual(potparam[(1, 1)]['epsilon'], 2.0)
        self.assertAlmostEqual(potparam[(1, 1)]['sigma'], 2.0)
        self.assertAlmostEqual(potparam[(1, 1)]['rcut'], 2.5)
        self.assertAlmostEqual(potparam[(0, 1)]['epsilon'], 1.4142135623730951)
        self.assertAlmostEqual(potparam[(0, 1)]['sigma'], 1.4142135623730951)
        self.assertAlmostEqual(potparam[(0, 1)]['rcut'], 2.5)

    def test_empty_create_potential(self):
        """Test the create potential method."""
        correct = {'potential': [{'class': 'Fake'}]}
        potentials, pot_par = create_potentials(correct)
        for i, j in zip(potentials, pot_par):
            self.assertIsNone(i)
            self.assertIsNone(j)

    def test_potential_inconsitentdim(self):
        """Test creation of potentials with inconsistent dims."""
        data = _read_raw_settings('settings-potential2.rst')
        correct = {'system': {'dimensions': 2},
                   'potential': [{'class': 'PairLennardJonesCut',
                                  'shift': True,
                                  'parameter': {0: {'sigma': 1.0,
                                                    'epsilon': 1.0,
                                                    'rcut': 2.5}}}]}
        settings = _test_correct_parsing(self, data, correct)
        args = [settings]
        self.assertRaises(ValueError, create_potentials, *args)

    def test_potential_create(self):
        """Test that we can create all potentials."""
        data = _read_raw_settings('allpotentials.rst')
        all_potentials = [PairLennardJonesCut,
                          PairLennardJonesCutnp,
                          DoubleWellWCA,
                          DoubleWell,
                          RectangularWell]
        raw = _parse_sections(data.split('\n'))
        settings = _parse_all_raw_sections(raw)
        potentials, _ = create_potentials(settings)
        for pot, pot_input in zip(potentials, all_potentials):
            self.assertIsInstance(pot, pot_input)

    def test_ext_potential(self):
        """Test creation of potentials while parsing input from externals."""
        data = _read_raw_settings('settings-ext-potential.rst')
        correct = {'potential': [{'class': 'FooPotential',
                                  'module': 'foopotential.py',
                                  'parameter': {'a': 2.0}}]}
        settings = _test_correct_parsing(self, data, correct)
        self.assertEqual(settings, correct)
        # Add path for testing:
        settings['simulation'] = {'exe_path': LOCAL_DIR}
        potentials, pot_param = create_potentials(settings)
        self.assertIsInstance(potentials[0], PotentialFunction)
        self.assertAlmostEqual(potentials[0].params['a'], 0.0)
        for pot, pot_param in zip(potentials, pot_param):
            pot.set_parameters(pot_param)
        self.assertAlmostEqual(potentials[0].params['a'], 2.0)

    def test_ext_potentialfail(self):
        """Test failure of external potential creation."""
        data = _read_raw_settings('settings-ext-potential-fail.rst')
        correct = {'potential': [{'class': 'BarPotential',
                                  'module': 'foopotential.py',
                                  'parameter': {'a': 2.0}}]}
        settings = _test_correct_parsing(self, data, correct)
        self.assertEqual(settings, correct)
        settings['simulation'] = {'exe_path': LOCAL_DIR}
        args = [settings]
        self.assertRaises(ValueError, create_potentials, *args)

    def test_complicated_input(self):
        """Test that we can read 'complex' force field input."""
        data = _read_raw_settings('forcefield.rst')
        correct = {'forcefield': {'description': 'My force field mix'},
                   'potential': [{'class': 'PairLennardJonesCutnp',
                                  'shift': True,
                                  'parameter': {0: {'sigma': 1.0,
                                                    'epsilon': 1.0,
                                                    'rcut': 2.5}}},
                                 {'class': 'DoubleWellWCA',
                                  'parameter': {'types': [(0, 0)],
                                                'rzero': 1. * (2.**(1./6.)),
                                                'height': 6.0, 'width': 0.25}},
                                 {'class': 'FooPotential',
                                  'module': 'foopotential.py',
                                  'parameter': {'a': 10.0}}]}
        settings = _test_correct_parsing(self, data, correct)
        self.assertEqual(settings, correct)
        settings['simulation'] = {'exe_path': LOCAL_DIR}
        forcefield = create_force_field(settings)
        self.assertEqual(len(forcefield.potential), 3)
        self.assertIsInstance(forcefield.potential[0], PairLennardJonesCutnp)
        self.assertIsInstance(forcefield.potential[1], DoubleWellWCA)
        self.assertIsInstance(forcefield.potential[2], PotentialFunction)

    def test_too_many(self):
        """Test what happens when we add too many potentials."""
        settings = {
            'forcefield': {'description': 'My force field mix'},
            'potential': [],
        }
        for i in range(66):
            settings['potential'].append(
                {'class': 'DoubleWellWCA',
                 'parameter': {'types': [(i, i)],
                               'rzero': 1. * (2.**(1./6.))}}
            )
        txt = settings_to_text(settings)
        raw = _parse_sections(txt.split('\n'))
        settings2 = _parse_all_raw_sections(raw)
        for i, pot in enumerate(settings2['potential']):
            self.assertEqual(pot['parameter']['types'][0], (i, i))

        # Looking for troubles:
        settings['ensemble'] = []
        for i in range(66):
            settings['ensemble'].append(
                {'potential': [{'class': 'DoubleWellWCA',
                                'parameter': {'types': [(i, i)],
                                              'rzero': 16}}]})

        txt = settings_to_text(settings)
        raw = _parse_sections(txt.split('\n'))
        settings3 = _parse_all_raw_sections(raw)
        hidden = settings3['ensemble'][10]['potential'][0]
        self.assertEqual(hidden['parameter']['rzero'], 16)

    def test_copy_settings(self):
        """Test that the copy method works as intended."""
        inputfile = os.path.join(LOCAL_DIR, 'settings-retis.rst')
        settings = parse_settings_file(inputfile)
        settings2 = copy_settings(settings)
        for key, val in settings.items():
            self.assertTrue(key in settings2)
            self.assertEqual(val, settings2[key])

    def test_settings_to_txt(self):
        """Test the settings to text in more detail."""
        inputfile = os.path.join(LOCAL_DIR, 'settings-retis.rst')
        settings = parse_settings_file(inputfile)
        txt = settings_to_text(settings)
        to_find = ['RETIS EXAMPLE FOR TESTING',
                   'TIS', 'RETIS']
        found = [False for _ in to_find]
        for lines in txt.split('\n'):
            for i, key in enumerate(to_find):
                if found[i]:
                    continue
                if lines.startswith(key):
                    found[i] = True
        self.assertTrue(all(found))

    def test_write_settings(self):
        """Test that we can write settings to a file."""
        inputfile = os.path.join(LOCAL_DIR, 'settings-retis.rst')
        settings = parse_settings_file(inputfile)
        with tempfile.TemporaryDirectory() as tempdir:
            out_file = os.path.join(tempdir, '_pyretis_settings_temp.rst')
            # Check that we can write a file:
            write_settings_file(settings, out_file)
            self.assertTrue(os.path.isfile(out_file))
            # Check that we can write a file and backup:
            write_settings_file(settings, out_file, backup=True)
            out_file2 = f'{out_file}_000'
            with open(out_file, 'r', encoding='utf-8') as fileh:
                raw_sections = _parse_sections(fileh)
            settings2 = _parse_all_raw_sections(raw_sections)
            self.assertTrue(os.path.isfile(out_file2))
        self.assertTrue(big_fat_comparer(settings, settings2, hard=True))

    def test_look_for_files(self):
        """Test that we look and find the files."""
        input_files = look_for_input_files(input_path=LOCAL_DIR,
                                           required_files={},
                                           extra_files=[])
        self.assertTrue(input_files == {})

        with turn_on_logging(), patch('sys.stdout', new=StringIO()):
            # Check that ferrari.41 doesn't exist.
            with self.assertLogs('pyretis.inout.settings',
                                 level='INFO'):
                look_for_input_files(input_path='.',
                                     required_files={},
                                     extra_files=['ferrari.f41'])

        # the index key should actually have an index.ndx value, but
        # I just use the available fake.mdp file instead...
        input_files = look_for_input_files(input_path=LOCAL_DIR,
                                           required_files={'mdp': 'gigi.mdp'},
                                           extra_files={'index': 'fake.mdp'})
        self.assertTrue('fake.mdp' in input_files['index'])

        input_files = look_for_input_files(input_path=LOCAL_DIR,
                                           required_files={'mdp': 'gigi.mdp'},
                                           extra_files=[])
        self.assertTrue('fake.mdp' in input_files['mdp'])
        input_files = look_for_input_files(input_path=LOCAL_DIR,
                                           required_files={'mdp': 'gigi.mdp'},
                                           extra_files=['traj.trr'])
        self.assertTrue('fake.mdp' in input_files['mdp'])

        input_files = look_for_input_files(input_path=LOCAL_DIR,
                                           required_files={'mdp': 'fake.mdp'})
        self.assertTrue('fake.mdp' in input_files['mdp'])

        input_files = look_for_input_files(input_path=LOCAL_DIR,
                                           required_files={'mdp': 'gigi.mdp',
                                                           'g96': 'dumped.g96',
                                                           'trr': 'traj.trr'})
        self.assertTrue('fake.mdp' in input_files['mdp'])
        self.assertTrue('dumped.g96' in input_files['g96'])
        self.assertTrue('traj.trr' in input_files['trr'])

        look_for_input_files(input_path=LOCAL_DIR,
                             required_files={'cf': 'config.gro'},
                             extra_files=[])
        with self.assertRaises(ValueError) as err:
            look_for_input_files(
                input_path=LOCAL_DIR,
                required_files={'lol': 'config.gRo'},
                extra_files=[])
        self.assertEqual('Missing input file "config.gRo" and multiple files '
                         + 'have extension ".gro"', str(err.exception))

        with self.assertRaises(ValueError) as err:
            input_files = look_for_input_files(input_path=LOCAL_DIR,
                                               required_files={'': 'ma.gro'})
        self.assertEqual('Missing input file "ma.gro" and multiple '
                         + 'files have extension ".gro"', str(err.exception))

        with self.assertRaises(ValueError) as err:
            input_files = look_for_input_files(input_path='banana',
                                               required_files={'': 'suka.gro'})
        self.assertEqual('Input path folder banana not existing',
                         str(err.exception))

        with self.assertRaises(ValueError) as err:
            input_files = look_for_input_files(
                input_path=LOCAL_DIR,
                required_files={'yyy': 'config.xxx'},
                extra_files=[])
        self.assertEqual('Missing input file "config.xxx" ',
                         str(err.exception))

    def test_clean_settings(self):
        """Test that we can clean settings."""
        inputfile = os.path.join(LOCAL_DIR, 'settings-retis.rst')
        settings = parse_settings_file(inputfile)
        settings['junk'] = {'some': 'junk'}
        settings['system']['junk'] = 'this is junk'
        settings['box'] = []
        settings_c = _clean_settings(settings)
        self.assertFalse('junk' in settings_c)
        self.assertFalse('junk' in settings_c['system'])
        self.assertFalse('box' in settings_c)


class AddDefaultTest(unittest.TestCase):
    """Test adding default values."""

    def test_add_default_settings(self):
        """Test that we can add default settings."""
        settings_base = SECTIONS
        add_default_settings(settings_base)
        self.assertTrue(big_fat_comparer(settings_base, SECTIONS))

    def test_add_specific_default_settings(self):
        """Test that we can add default settings."""
        settings = {}
        add_default_settings(settings)

        settings['simulation']['task'] = 'tis'
        add_specific_default_settings(settings)
        self.assertEqual(settings['system']['temperature'], 1)
        self.assertFalse(settings['simulation']['flux'])
        self.assertFalse(settings['simulation']['zero_ensemble'])

        settings['simulation']['task'] = 'retis'
        del settings['simulation']['flux']
        del settings['simulation']['zero_ensemble']
        add_specific_default_settings(settings)
        self.assertTrue(settings['simulation']['flux'])
        self.assertTrue(settings['simulation']['zero_ensemble'])

        self.assertEqual(settings['particles']['type'], 'internal')
        self.assertEqual(settings['engine']['type'], 'internal')

        settings['engine']['class'] = 'cp2k'
        settings['engine']['exe_path'] += '/test/inout'
        add_specific_default_settings(settings)
        self.assertEqual(settings['particles']['type'], 'external')
        self.assertEqual(settings['engine']['type'], 'external')
        input_path = settings['engine'].get('input_path', '.')
        add_specific_default_settings(settings)
        self.assertEqual(settings['engine']['input_files']['conf'][-11:],
                         'initial.xyz')
        self.assertEqual(settings['engine']['input_files']['template'][-9:],
                         '/cp2k.inp')
        self.assertEqual(
            settings['engine']['input_files']['template'],
            os.path.join(os.path.abspath('.') + '/test/inout/',
                         input_path, 'cp2k.inp'))
        self.assertEqual(settings['system']['temperature'], 500)
        del settings['system']['temperature']
        add_specific_default_settings(settings)
        self.assertEqual(settings['system']['temperature'], 500)
        no_temp_inp = os.path.join(os.path.abspath('.') + '/test/inout/',
                                   input_path, 'no_temp.inp')
        settings['engine']['template'] = no_temp_inp
        with turn_on_logging(), patch('sys.stdout', new=StringIO()):
            with self.assertLogs() as log:
                del settings['system']['temperature']
                add_specific_default_settings(settings)
                self.assertEqual(settings['system']['temperature'], 300)
                self.assertIn('Temperature not set in CP2K', log.output[0])
                settings['system']['temperature'] = 500
                add_specific_default_settings(settings)
                self.assertEqual(settings['system']['temperature'], 300)
                self.assertIn('And temperature in input', log.output[1])

        settings = {}
        settings['simulation'] = {'covid': 'kill_us_all'}
        settings['initial-path'] = {'method': 'restart'}
        add_default_settings(settings)
        settings['engine']['class'] = 'gromacs2'
        settings['engine']['topology'] = 'TrainToRide'
        add_specific_default_settings(settings)
        self.assertEqual(settings['particles']['type'], 'external')
        self.assertEqual(settings['engine']['type'], 'external')
        input_path = settings['engine'].get('input_path', '.')
        self.assertEqual(settings['engine']['input_files']['topology'],
                         'TrainToRide')
        self.assertEqual(settings['engine']['input_files']['conf'][-8:],
                         'conf.gro')
        self.assertEqual(settings['engine']['input_files']['input_o'][-11:],
                         '/grompp.mdp')
        self.assertEqual(settings['engine']['input_files']['index'],
                         os.path.join(os.path.abspath('.'),
                                      input_path, 'index.ndx'))
        self.assertEqual(settings['simulation']['restart'], 'pyretis.restart')


class KeywordEnsemble(unittest.TestCase):
    """Test ensemble and setting keywords."""

    def test_fill_up_tis_and_retis_settings(self):
        """Test if ensembles are correctly generated."""
        inputfile = os.path.join(LOCAL_DIR, 'settings-retis.rst')
        settings = parse_settings_file(inputfile)

        self.assertEqual(settings['ensemble'][2]
                         ['collective-variable'][3]['something']
                         ['unexpected'], [1, 2, 3])
        self.assertEqual(settings['ensemble'][6]['particles']
                         ['velocity']['fantasy'], 'game')
        self.assertEqual(settings['ensemble'][6]['particles']['velocity']
                         ['generate'], 'Priapo')
        self.assertEqual(settings['ensemble'][5]['retis'], settings['retis'])
        self.assertEqual(settings['ensemble'][2]['box'], settings['box'])
        self.assertTrue(settings['ensemble'][0]['tis']['ensemble_number'] !=
                        settings['ensemble'][1]['tis']['ensemble_number'])
        self.assertEqual(settings['ensemble'][2]
                         ['collective-variable'][5]['name'], 'Bugno')
        self.assertEqual(settings['ensemble'][2]
                         ['collective-variable'][1]
                         ['position']['nonsense'], 'pineapple_on_pizza')
        self.assertEqual(settings['ensemble'][1]['particles']
                         ['velocity']['generate'], 'maxwell')
        self.assertEqual(len(settings['ensemble']), 8)
        self.assertEqual(settings['simulation']['zero_left'], -99)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Example of creating a lattice and making a box."""
from pyretis.core import create_box
box1 = create_box()
print(box1)
box2 = create_box(cell=[10, 10, 10])
print(box2)
box3 = create_box(low=[0, -10, 10], high=[10, 10, 20],
                  periodic=[True, True, False])
print(box3)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Compare the energy output for GROMACS and PyRETIS."""
import sys
import colorama
from matplotlib import pyplot as plt
import numpy as np
from pyretis.inout import print_to_screen
from pyretis.inout.formats.gromacs import (
    read_xvg_file,
)


plt.style.use('seaborn-v0_8-poster')


def main(energy_file, xvg_file, plot=False):
    """Perform the test."""
    print_to_screen(f'Reading energy file: {energy_file}',
                    level='info')
    energy = np.loadtxt(energy_file)
    print_to_screen(f'Reading xvg file: {xvg_file}', level='info')
    energy_xvg = read_xvg_file(xvg_file)

    mse_ok = obtain_mses(energy, energy_xvg)

    if plot:
        print_to_screen('\nPlotting for comparison', level='message')
        plot_comparison(energy, energy_xvg, energy_file, xvg_file)

    if not mse_ok:
        print_to_screen('\nComparison failed!', level='error')
        sys.exit(1)


def obtain_mses(energy, energy_xvg, tol=1.0e-5):
    """Obtain some mean squared errors."""
    pairs = ((1, 'potential'), (2, 'kinetic en.'), (3, 'total energy'),
             (4, 'temperature'))
    for pair in pairs:
        if not len(energy[:, pair[0]]) == len(energy_xvg[pair[1]]):
            return False
        mse = ((energy[:, pair[0]] - energy_xvg[pair[1]])**2).mean(axis=0)
        level = 'info'
        tol_ok = True
        if tol:
            tol_ok = abs(mse) < tol
            if not tol_ok:
                level = 'error'
        print_to_screen(f'MSE {pair[1]}: {mse}', level=level)
        if not tol_ok:
            return False
    return True


def plot_comparison(energy, energy_xvg, energy_file, xvg_file):
    """Just plot some properties for the paths."""
    fig1 = plt.figure()
    ax11 = fig1.add_subplot(221)
    ax12 = fig1.add_subplot(222)
    ax21 = fig1.add_subplot(223)
    ax22 = fig1.add_subplot(224)
    plabel = f'PyRETIS ({energy_file})'
    glabel = f'gmx ({xvg_file})'
    ax11.plot(energy[:, 1], lw=4, ls='-', marker='o', label=plabel)
    ax11.plot(energy_xvg['potential'], lw=2, ls='--', marker='^',
              label=glabel)
    ax11.legend()
    ax11.set_ylabel('Potential')

    ax12.plot(energy[:, 2], lw=4, ls='-', marker='o')
    ax12.plot(energy_xvg['kinetic en.'], lw=2, ls='--', marker='^')
    ax12.set_ylabel('Kinetic')

    ax21.plot(energy[:, 3], lw=4, ls='-', marker='o')
    ax21.plot(energy_xvg['total energy'], lw=2, ls='--', marker='^')
    ax21.set_ylabel('Total')

    ax22.plot(energy[:, 4], lw=4, ls='-', marker='o')
    ax22.plot(energy_xvg['temperature'], lw=2, ls='--', marker='^')
    ax22.set_ylabel('Temperature')

    fig1.tight_layout()
    plt.show()


if __name__ == '__main__':
    colorama.init(autoreset=True)
    PLOT = len(sys.argv) > 3
    main(sys.argv[1], sys.argv[2], plot=PLOT)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""GUI application for visualizing simulation data.

This is a PyQt5 file, using a custom made ui layout, with a window which
displays different plots created by the path density tool. Window can either
display data of a pre-compiled compressed file generated with the
orderparam_density.py module, or compile on in/out files by importing
orderparam_density and executing before displaying the results.


Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

visualize_main (:py:func: `.visualize_main`)
    Method to load the GUI.


Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

VisualApp (:py:class:`.VisualApp`)
    The PyQt5 GUI class that handles all the functionalities.

CustomFigCanvas (:py:class:`.VisualApp`)
    Class to handle the MatPlotLiv canvas.

DataSlave (:py:class:`.VisualApp`)
    The class to handle the data for the plots.

VisualObject (:py:class:`.VisualApp`)
    The class that handles the load from files.

DataObject (:py:class:`.VisualApp`)
    The class that loads the PyRETIS files.

"""
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
import os
import warnings
import subprocess
import sys
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas
)
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import mdtraj as md
import numpy as np
import pandas as pd
from PyQt5 import QtCore, QtGui
from PyQt5 import QtWidgets, uic
from scipy.spatial import distance
from pyretis.inout import settings
from pyretis.inout.common import TRJ_FORMATS
from pyretis.setup.common import create_orderparameter
from pyretis.tools.recalculate_order import recalculate_order
from pyretis.pyvisa.common import (read_traj_txt_file,
                                   recalculate_all,
                                   try_data_shift,
                                   find_rst_file,
                                   get_cv_names)
from pyretis.pyvisa.orderparam_density import (PathDensity,
                                               PathVisualize,
                                               Trajectory,
                                               pyvisa_zip,
                                               remove_nan)
from pyretis.pyvisa.plotting import (gen_surface,
                                     plot_regline,
                                     plot_int_plane)
from pyretis.pyvisa.statistical_methods import (pyvisa_pca,
                                                k_means,
                                                spectral,
                                                gaussian_mixture,
                                                hierarchical,
                                                correlation_matrix,
                                                random_forest,
                                                decision_tree)

warnings.filterwarnings('ignore', category=pd.io.pytables.PerformanceWarning)

# Hard-coded labels for energies and time/cycle steps
ENERGYLABELS = ['time', 'cycle', 'potE', 'kinE', 'totE']


UI_VW, QT_BC = uic.loadUiType(
    os.path.dirname(os.path.realpath(__file__)) +
    '/' + 'pyvisaGUI.ui')


class VisualApp(QtWidgets.QMainWindow, UI_VW):
    """Class definition of the path visualization window GUI for PyRETIS.

    Application opens a QMainWindow object with a QFrame for inserting
    a matplotlib figure canvas into, and a QDropWidget with the relevant
    settings for plotting.

    Attributes
    ----------
    mainworker : A QThread object that stores (and executes) the path density
        data back-end, initialized by the VisualApp window. When
        called will update the data lists and return to VisualApp
        for plotting.

    Functions
    ---------
    close_event, action_reload, center_on_screen, toggle_buttons :
        Functions that relate to closing of window, reloading data,
        centering the window on screen and disable key buttons while
        pdating data/figure.
    _update_canvas_{title/text} :
        Functions that update the title and label text of plot settings
        and in the figure.
    _get_settings : Gets plot settings from VisualApp dropwidget.
    _get_savename : Generate generic savename based on current settings.
    save_png : Function that store the current figure shown in VisualApp
        window as a .png image, with name generated from data settings.
    save_hdf5 : Function that store the current figure shown in VisualApp
        window a hdf5 save, with name generated from data settings.
    save_json : Function that store the current figure shown in VisualApp
        window a json save, with name generated from data settings.
    save_script : Function that generates a script to reconstruct the current
        plot from the compressed file.
    _load_file : Opens QFileDialog for choosing file to load in VisualApp.
    _load_data : Loads data from hdf5 save.
    _load_data_output : Executes a PathDensity on input/output file before
        displaying results.
    _reload : Clears old data and initializes loading new data from file.
    _change_{cmap/zoom} :
        Updates figure with different colormap, and with xlims+ylims
    toggle_{intf/regl/cbar/titles} :
        Functions that shows/hides interfaces, regression line,
        colorbar, and titles/labes, respectively.
    emit_settings : Function that is called when pressing a QPushButton labeled
        'Update' after choosing desired settings for plot.
        Sends settings to mainworker, which updates the data of
        x, y (and z) lists before sending back to VisualApp for
        plotting.
    on_pick : Function that is called when clicking on the plot.
        Writes the x, y coordinate, cycle, ensemble name,
        status and mc move of the closest corresponding point to
        wherever the user clicks on the plot and fills in the slots
        in the Analysis tab in PyVisA and plots the closest point.
    draw_trajectory : Function that is called when pressing a QPushButton.
        When pressing the button labeled "show trj" in the "Analysis"
        tab in PyVisA. The function plots all points in the trajectory
        corresponding to the closest point defined by the
        users pick on the plot.
    play_cycle : Function that is called when pressing a QPushButton.
        Pressing the QPushButton labeled "play" button in the "Analysis"
        tab in PyVisA causes the function to check if the corresponding
        trajectory files exists, and reverses them through rev_trj if
        necessary before merging them to a pdb file, and loading them
        to PyMOL. NB: Requires the user to have PyMOL installed in
        order to work.
    create_full_traj: Function that creates a pdb-file for a trajectory.
    display_message_box : Function that displays a QMessageBox.
        Function displays a message box in use for error handling and
        messaging the user about operations and choices that are not
        availeable.
    toggle_only_stored_trjs: Only plot data with trajectory files.
        Function that loops over all trajectories in the chosen ensembles
        and cycles and only plots those trajectories which have
        existing trajectory files.
    recalculate : Recalculate OP and CV's in new order.txt files.
        Function activating from the QAction labeled
        "Recalculate data" in the file menu of PyVisA. Uses the tools
        available in the recalculate_order.py file to load the
        trajectory files of the simulation, and calculate new
        collective variables which need to be declared in the
        input files "retis.rst" and "orderp.py". The new data will be
        stored in the full order.txt in the ensemble folders, and
        also the order.txt in the cycle folders of each ensemble
        enabling PyVisA to load and visualize the new descriptors.
        The previous data files are renamed as order_old.
    statistic_analysis: Perform statistical analysis on plotted data.
        A function that is connected to the statistical function in the
        statistical_methods file, and that takes the users choice of analysis
        and run the said choice.
    show_correlation : Show the correlation between simulation variables.
        Shows the Pearson correlation in a matrix form for all the variables
        from the simulation.
    refresh_data: Signal a walk of the data.
        Performs a new walk of the data, to use during a running simulation
        to visualise its development.

    Signals
    -------
    start_cmd : PyQt Signal
        It tells worker to start using string, cmd.
    send_settings : PyQt Signal
        It sends settings for data_get to worker thread.

    Slots
    -----
    update_cycle : function bound to a pyqtSignal sent by the mainworker
        thread. Signal sets upper and lower bounds for max/min cycles used
        in plotting.
    update_fig : PyQt Slot
        Pyqt slot, worker thread calls function by sending 3 lists
        of floats for plotting in this function.

    """

    send_settings = QtCore.pyqtSignal(dict, name='Data settings')
    start_cmd = QtCore.pyqtSignal(str)

    def __init__(self, folder, iofile, sub_files):
        """Initialize the VisualApp QMainWindow.

        Sets up ui from file, centers on screen and sets up mainworker QThread
        for data loading.

        Parameters
        ----------
        folder : string
            The path of the target folder.
        iofile : string
            The name of the main input file.
        sub_files : dict
             Contains the name of multiple input files.

        """
        QtWidgets.QMainWindow.__init__(self)
        UI_VW.__init__(self)
        # Delete memory on close?
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.setupUi(self)
        self.center_on_screen()
        self.setWindowTitle('PyVisA - Visualization and Analysis')
        self.folder = folder
        self.iofile = iofile

        # Link to the input file if PyVisA was loaded with a rst file
        self.rst_file = sub_files['rst']
        # Link to the trajectory file if -data was used
        self.trajfile = sub_files['traj']
        self.settings, self.dataobject, self.thread = None, None, None
        # Load data in separate thread
        if iofile is not None:
            self._load_file()
        self._init_widget()
        self.myfig.fig.canvas.mpl_connect(
            'button_press_event', self.on_pick)

        self.xmin, self.xmax = None, None
        self.ymin, self.ymax = None, None
        self.zmin, self.zmax = None, None

    def close_event(self):
        """Event function interpreter to make the event to a boolean."""
        self.close()

    def closeEvent(self, event):  # pylint: disable=C0103
        """Event function, activated when attempting to close VisualApp.

        Will create a QMessage prompt asking for confirmation of exit by
        user. If confirmed, will stop Qthread of DataObject and attempt to
        clear dataobject from memory.

        Parameters
        ----------
        event : QtWidget event
            It activates the closing of the VisualApp.

        """
        reply = QtWidgets.QMessageBox.question(
            self, 'Quit?', 'Are you sure you want to quit?',
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.No)

        if reply == QtWidgets.QMessageBox.Yes:
            self.statusbar.showMessage('Closing window')
            try:
                self.thread.quit()
                self.thread.wait()
                del self.dataobject
                del self.thread
            except AttributeError:
                pass
            finally:
                event.accept()
        else:
            event.ignore()

    def action_reload(self):
        """Display a QMessagebox to confirm reload action.

        Function that displays a QMessagebox for user, double confirming
        the reload data action.
        """
        if not self.dataobject:
            self._reload()
        else:
            load = QtWidgets.QMessageBox.question(
                self,
                "Reload data from file",
                "Are you sure to discard the current data and "
                "reload a new set ?",
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
            if load == QtWidgets.QMessageBox.Yes:
                self.statusbar.showMessage('Deleting old data')
                try:
                    self.thread.quit()
                    self.thread.wait()
                except AttributeError:
                    # No self.thread running.
                    pass
                self._reload()

    def center_on_screen(self):
        """Centers widget/window on screen."""
        resolution = QtWidgets.QDesktopWidget().screenGeometry()
        self.move(int((resolution.width() - self.frameSize().width()) / 2),
                  int((resolution.height() - self.frameSize().height()) / 2))

    def toggle_buttons(self, onoff):
        """Toggles enable-state of key buttons of VisualApp.

        It avoids user-created conflicts while busy updating data and drawing
        figure.

        Parameters
        ----------
        onoff : boolean
            Determines the status of the button.

        """
        self.updateBtn.setEnabled(onoff)
        self.previewBtn.setEnabled(onoff)
        self.intShowChkBtn.setEnabled(onoff)
        self.regLineChkBtn.setEnabled(onoff)
        self.showTitleChkBtn.setEnabled(onoff)
        self.saveFigBtn.setEnabled(onoff)
        self.previewBtn.setEnabled(onoff)
        self.refreshBtn.setEnabled(onoff)

    def disable_zcombox(self):
        """Disabling combobox with z-values if plotting density."""
        if 'Density' in self.plotTypeComBox.currentText():
            self.energyComBox.setEnabled(False)
        else:
            self.energyComBox.setEnabled(True)

    def disable_reglinecheck(self):
        """Disabling regression line checkbox if 3D method."""
        if '3D' in self.plotTypeComBox.currentText():
            self.regLineChkBtn.setEnabled(False)
        elif 'Contour' in self.plotTypeComBox.currentText():
            self.regLineChkBtn.setEnabled(False)
        elif 'Surface' in self.plotTypeComBox.currentText():
            self.regLineChkBtn.setEnabled(False)
        else:
            self.regLineChkBtn.setEnabled(True)

    def _update_canvas_font(self):
        """Change the fontsize of title/labels of figure.

        Bound to event only.
        """
        titlefont = self.titleSizeSpin.value()
        self.settings['title font'] = titlefont
        axesfont = self.axesSizeSpin.value()
        self.settings['axes font'] = axesfont
        self.myfig.title.set_fontsize(titlefont)
        self.myfig.xaxis.set_fontsize(axesfont)
        self.myfig.yaxis.set_fontsize(axesfont)
        self.myfig.zaxis.set_fontsize(axesfont)
        self.myfig.cbar.ax.tick_params(labelsize=axesfont)
        for tick in self.myfig.ax.xaxis.get_major_ticks():
            tick.label.set_fontsize(axesfont)
        for tick in self.myfig.ax.yaxis.get_major_ticks():
            tick.label.set_fontsize(axesfont)
        # Default pyplot tick size: lenght=3.5, width=1.0
        tick_l = round(axesfont / 4, 2)
        tick_w = round(axesfont / 10, 2)
        self.myfig.ax.tick_params(length=tick_l, width=tick_w)
        self.myfig.fig.canvas.draw()

    def _update_canvas_text(self):
        """Create the titles and labels of figure.

        This is in VisualApp from QLineEdits current text.
        Called by update_fig() and bound to event.
        """
        self.settings['show titles'] = self.showTitleChkBtn.isChecked()
        show = self.settings['show titles']
        self.myfig.title = self.myfig.ax.set_title(
            self.titleLine.text(),
            fontsize=self.settings['title font'],
            visible=show)
        self.myfig.xaxis = self.myfig.ax.set_xlabel(
            self.xaxisLabel.text(),
            fontsize=self.settings['axes font'],
            visible=show)
        self.myfig.yaxis = self.myfig.ax.set_ylabel(
            self.yaxisLabel.text(),
            fontsize=self.settings['axes font'],
            visible=show)
        self.myfig.zaxis = self.myfig.cbar_ax.set_title(
            self.zaxisLabel.text(),
            fontsize=self.settings['axes font'],
            visible=show)
        self.myfig.fig.canvas.draw()

    def _get_titles(self):
        """Generate tiles and labels and update QLineEdits of VisualApp.

        Function that generates generic titles and labels for the figure
        axes, colorbar, etc. and updates QLineEdits of VisualApp with
        those titles/labels.
        """
        method = self.settings['method']
        name = method[0]
        acc = self.settings['ACC']
        shift = self.settings['try_shift']
        fol = self.settings['fol']
        op1 = self.settings['op1']
        op2 = self.settings['op2']
        energy = self.settings['E']

        xyz = op1 + '-' + op2
        if 'Density' not in method:
            xyz += '-' + energy
            self.myfig.cbar_ax.set_title(energy)
        if self.settings['weight']:
            name += '(W)'
        cmin, cmax = self.settings['min_cycle'], self.settings['max_cycle']
        title = (f'{name}, using {fol} folder(s), with {acc} paths, '
                 f'cycles: {cmin}-{cmax}')
        if shift:
            title += '\nData shifted'
        # Setting text in QLineEdits of VisualApp
        self.xaxisLabel.setText(op1)
        self.yaxisLabel.setText(op2)
        if self.settings['method'][0] == 'Density':
            self.zaxisLabel.setText('Density')
        else:
            self.zaxisLabel.setText(energy)
        self.titleLine.setText(title)
        # Updating text in plot from text in QLineEdits
        self._update_canvas_text()

    def _get_savename(self):
        """Generate generic save-name.

        Function that generates a generic save-name for the current
        figure displayed in VisualApp based on settings.
        """
        method = self.settings['method']
        method_name = method[0]
        acc = self.settings['ACC']
        fol = self.settings['fol']
        op1 = self.settings['op1']
        op2 = self.settings['op2']
        xyz = op1 + '-' + op2
        cmin, cmax = self.settings['min_cycle'], self.settings['max_cycle']
        if 'Density' not in method and self.settings['E'] != 'None':
            xyz += '-' + self.settings['E']
        elif self.settings['weight']:
            method_name += '(W)'
        name = f'{method_name}_{xyz}_ACC={acc}_folder={fol}_({cmin}-{cmax})'
        return name

    @staticmethod
    def _iffloat(value, default_v):
        """Check if value is a float, and returns it."""
        try:
            return float(value)
        except ValueError:
            return default_v

    def _set_def_limits(self, x, y, z):
        """Update x/y/z limits in GUI with default values.

        Parameters
        ----------
        x, y, z : list
            Floats, the coordinates used in plotting.

        """
        self.statusbar.showMessage('Find limits of data selected.')

        self.xmin = min(x, default=0)
        self.ymin = min(y, default=0)
        self.zmin = min(z, default=0)
        self.xmax = max(x, default=0)
        self.ymax = max(y, default=0)
        self.zmax = max(z, default=0)

        self.xminLine.setText(str(self.xmin))
        self.xmaxLine.setText(str(self.xmax))
        self.yminLine.setText(str(self.ymin))
        self.ymaxLine.setText(str(self.ymax))
        self.zminLine.setText(str(self.zmin))
        self.zmaxLine.setText(str(self.zmax))

    def _set_cur_limits(self):
        """Update x/y/z limits in GUI with current values."""
        self.xminLine.setText(str(self.settings['x-limits'][0]))
        self.yminLine.setText(str(self.settings['y-limits'][0]))
        self.zminLine.setText(str(self.settings['z-limits'][0]))
        self.xmaxLine.setText(str(self.settings['x-limits'][1]))
        self.ymaxLine.setText(str(self.settings['y-limits'][1]))
        self.zmaxLine.setText(str(self.settings['z-limits'][1]))

    def _get_limits(self):
        """Update x/y/z limits from GUI.

        Function that updates the min/max limits of data for plotting
        in the GUI. If invalid inputs, will return default values.
        """
        try:
            float(self.xmin)
        except (AttributeError, TypeError):
            # Not yet defined data, skip
            return

        # Check min/max of x
        nxi = self._iffloat(self.xminLine.text(), self.xmin)
        nxa = self._iffloat(self.xmaxLine.text(), self.xmax)
        if nxi >= nxa:
            self.settings['x-limits'] = (self.xmin, self.xmax)
        else:
            self.settings['x-limits'] = (nxi, nxa)

        # Check min/max of y
        nyi = self._iffloat(self.yminLine.text(), self.ymin)
        nya = self._iffloat(self.ymaxLine.text(), self.ymax)
        if nyi >= nya:
            self.settings['y-limits'] = (self.ymin, self.ymax)
        else:
            self.settings['y-limits'] = (nyi, nya)

        # Check min/max of z
        nzi = self._iffloat(self.zminLine.text(), self.zmin)
        nza = self._iffloat(self.zmaxLine.text(), self.zmax)
        if nzi >= nza:
            self.settings['z-limits'] = (self.zmin, self.zmax)
        else:
            self.settings['z-limits'] = (nzi, nza)
        self._set_cur_limits()

    def _get_settings(self):
        """Update self.settings{} with the current option from the GUI.

        Function that updates the settings of dictionary
        self.settings{} with current options from GUI window.
        """
        full_method = self.plotTypeComBox.currentText()
        dim = int(full_method[0])
        method = full_method.split(' ')[1:]
        weight = bool(method[-1] == '(weight)')
        self.settings = {
            'op1': self.firstOpComBox.currentText(),
            'op2': self.secondOpComBox.currentText(),
            'E': self.energyComBox.currentText(),
            'ACC': self.accComBox.currentText(),
            'fol': self.folderComBox.currentText(),
            'min_cycle': self.minCyclSpin.value(),
            'max_cycle': self.maxCyclSpin.value(),
            'weight': weight,
            'method': method,
            'dim': dim,
            'try_shift': self.dataShiftChkBtn.isChecked(),
            'show_int': self.intShowChkBtn.isChecked(),
            'reg_line': self.regLineChkBtn.isChecked(),
            'res': self.resSpinBox.value(),
            'colormap': self.cmapComBox.currentText(),
            'colorbar': self.cbarChkBtn.isChecked(),
            'title font': self.titleSizeSpin.value(),
            'axes font': self.axesSizeSpin.value(),
            'show titles': self.showTitleChkBtn.isChecked(),
            'stored': self.storedTrjsChkBtn.isChecked(),
            'MC-move': self.mcComBox.currentText(),
            'start_end': self.start_end_box.currentText()
        }
        # Get limits of x/y/z data:
        self._get_limits()

    def save_hdf5(self):
        """Save in hdf5 format the current figure.

        Function that saves the current pyplot.figure object of
        VisualApp and CustomFigCanvas to a .hdf5 file. Function creates
        title and name for file by the current settings for the plot.
        """
        if self.settings is None:
            self.statusbar.showMessage('No data selected')
            return
        x, y, z = self.dataobject.load_all()
        dataframe = pd.Series({'x': x, 'y': y, 'z': z,
                               'settings': self.settings})
        name = self._get_savename() + '.hdf5'
        outfile = os.path.join(self.folder, name)
        dataframe.to_hdf(outfile, key='hdf5', mode='w')
        self.statusbar.showMessage(f'Figure saved as {outfile}')

    def save_json(self):
        """Save in json format the current figure.

        Function that saves the current pyplot.figure object of
        VisualApp and CustomFigCanvas to a .json file. Function creates
        title and name for file by the current settings for the plot.
        """
        if self.settings is None:
            self.statusbar.showMessage('No data selected')
            return
        name = self._get_savename() + '.json'
        outfile = os.path.join(self.folder, name)
        x, y, z = self.dataobject.load_all()
        dataframe = pd.Series({'x': x, 'y': y, 'z': z,
                               'settings': self.settings})
        dataframe.to_json(outfile)
        self.statusbar.showMessage(f'Figure saved as {outfile}')

    def save_png(self):
        """Save current figure as png.

        Function that saves the current plotted figure of VisualApp
        and CustomFigCanvas as a .png file.
        Function creates title and name for file by the current settings
        for the plot.
        """
        if self.settings is None:
            self.statusbar.showMessage('No data selected')
            return
        name = self._get_savename() + '.png'
        outfile = os.path.join(self.folder, name)
        self.myfig.fig.savefig(outfile)
        self.statusbar.showMessage(f'Figure saved: {outfile}')

    def save_textdata(self):
        """Save x, y, z data to file.

        Function that saves the current x, y, z data of the visualizer to
        a text file for use in other operations, plotting, etc.
        """
        if self.settings is None:
            self.statusbar.showMessage('No data selected')
            return
        self.statusbar.showMessage('Saving figure data to file...')
        name = self._get_savename() + '.txt'
        outfile = os.path.join(self.folder, name)
        x, y, z = self.dataobject.load_all()

        if not z:
            line = '\t{}\t\t{}\n'
            header = line.format(self.settings['op1'],
                                 self.settings['op2'])
        else:
            line = '\t{}\t\t{}\t\t{}\n'
            header = line.format(self.settings['op1'],
                                 self.settings['op2'],
                                 self.settings['E'])

        self.statusbar.showMessage(f'Writing file: {outfile}')
        with open(os.path.join(self.folder, outfile),
                  'w', encoding="utf8") as output:
            output.write(header)
            if not z:
                for i, j in zip(x, y):
                    output.write(line.format(i, j))
            else:
                for i, j, k in zip(x, y, z):
                    output.write(line.format(i, j, k))
        self.statusbar.showMessage(f'Figure data saved: {outfile}!')

    def save_script(self):
        """Make makefigure.py script.

        Function that generate and store the current plot from the GUI to
        a separate script file, that can re-generate a fairly similar
        figure to the GUI using stored settings.
        """
        if self.settings is None:
            self.statusbar.showMessage('No data selected')
            return
        datafile = "pyvisa_compressed_data.hdf5"
        self._save_sim_data_hdf5()
        scriptfileformat = 'makefigure_{}_{}_{}_{}.py'
        self._get_settings()
        # Text writes to makefigure.py
        txt = "# Makefigure script\n"
        txt += "import pandas as pd\n"
        txt += "import numpy as np\n"
        txt += "import os.path\n"
        txt += "import matplotlib.pyplot as plt\n"
        txt += "from scipy.interpolate import griddata as scgriddata\n"
        txt += "\n"
        txt += f"datafile = '{datafile}'\n"
        txt += "\n"
        txt += "traj = pd.read_hdf(datafile, key='data')\n"
        txt += "infos = pd.read_hdf(datafile, key='infos')\n"
        txt += "data = [traj, infos]\n"
        txt += "\n"
        txt += "# Dictionary with settings for data load and plotting:\n"
        txt += f"settings = {self.settings}\n"
        txt += "\n"

        if self.settings['fol'] == 'All':
            fol = []
            for folder in next(os.walk(self.folder))[1]:
                if folder[0] == '0':
                    fol.append(folder)
        else:
            fol = [self.settings['fol']]
        txt += f"fol = {fol}\n"
        x, y, z = self.dataobject.load_all()
        op1 = self.settings['op1']
        op2 = self.settings['op2']
        energy = self.settings['E']
        txt += f"xl = '{op1}'\n"
        txt += f"yl = '{op2}'\n"
        if self.settings['method'][0] == 'Density':
            index_data = 0
        else:
            index_data = 1
            txt += f"zl = '{energy}'\n"
        txt += "\n"
        txt += "fig = plt.figure()\n"
        txt += "ax = fig.add_subplot(111)\n"
        txt += "fig.subplots_adjust(left=0.1, right=0.85, "
        txt += "bottom=0.1, top=0.9)\n"
        txt += "cbar_ax = fig.add_axes([0.86, 0.1, 0.03, 0.8])\n"
        txt += "\n"
        txt += "# Get x, y, z data\n"
        txt += f"x, y, z = {x}, {y}, {z}\n"
        res = self.settings['res']
        cmap = self.settings['colormap']
        if index_data == 1:
            txt += f"xi = np.linspace(min(x), max(x), {res})\n"
            txt += f"yi = np.linspace(min(y), max(y), {res})\n"
            txt += "X, Y = np.meshgrid(xi, yi)\n"
            txt += "Z = scgriddata((x, y), np.array(z), (X, Y), "
            txt += "method='linear', fill_value=max(z))\n"
            txt += "\n"
            txt += f"surf = ax.contourf(X,Y,Z, cmap='{cmap}')\n"
            txt += "cbar = fig.colorbar(surf, cax=cbar_ax)\n"
        else:
            txt += f"surf = ax.hist2d(x, y, bins={res},"
            txt += f" cmap='{cmap}', density=True)\n"
            txt += "cbar = fig.colorbar(surf[3], cax=cbar_ax)\n"

        txt += "plt.show()"

        if index_data == 0:
            zlabel = 'Density'
        else:
            zlabel = energy
        scriptfile = scriptfileformat.format(
            op1, op2, zlabel, self.settings['fol'])
        with open(os.path.join(self.folder, scriptfile), 'w',
                  encoding="utf8") as output:
            output.write(txt)
        self.statusbar.showMessage(f'Script file saved: {scriptfile}')

    def _load_file(self):
        """Load data file.

        Function that sets up QObject, moves to QThread, and begins the
        data extract. Or, loads pre-processed data from file and
        moves to a QThread.
        """
        if self.folder is None and self.iofile is None:
            # When VisualApp is called without a directory, opens a filedialog
            iofile = QtWidgets.QFileDialog.getOpenFileName(
                parent=self,
                caption="Select input/output file in simulation directory")
            self.iofile = iofile[0]
            folder = os.path.dirname(os.path.realpath(self.iofile))
            self.folder = folder
            os.chdir(self.folder)
        # Setting name of file to QLineEdit of VisualApp
        if self.rst_file:
            self.filenameLine.setText(self.rst_file)
            self._load_data_output()
            return

        self.filenameLine.setText(self.iofile)
        # If PyVisA-data has been loaded
        if self.iofile.endswith(tuple(TRJ_FORMATS)) or \
                self.trajfile and \
                self.trajfile.endswith(tuple(TRJ_FORMATS)):
            self._load_data(self.iofile)
        elif self.iofile.endswith('.rst'):
            self._load_data_output()
        elif self.iofile.endswith(('.hdf5', '.zip')):
            self._load_data(self.iofile)
        else:
            msg = f'Format Error, file {self.iofile} not recognized.'
            self.statusbar.showMessage(msg)

    def _save_sim_data_hdf5_z(self):
        """Save the data to hdf5 zipped file."""
        self._save_sim_data_hdf5()
        file_o = 'pyvisa_compressed_data.hdf5'
        pyvisa_zip(file_o)

    def _save_sim_data_hdf5(self):
        """Save the data to hdf5 file."""
        file_o = 'pyvisa_compressed_data.hdf5'
        self.statusbar.showMessage(f'Saving data to {file_o}')
        data = pd.Series(self.dataobject.traj_data)
        infos = pd.Series(self.dataobject.infos)
        data.to_hdf(file_o, key='data', mode='w')
        infos.to_hdf(file_o, key='infos')
        self.statusbar.showMessage(f'Figure saved as {file_o}')

    def _load_data_output(self):
        """Load simulation data.

        Function that loads simulation data using in/out file of simulation
        and initializes PathDensity on the directory, before showing the
        results in the VisualApp window.
        """
        self.statusbar.showMessage('Loading data from output files')
        # Set-up data object using PathDensity class and moving to thread
        self.thread = QtCore.QThread()
        self.dataobject = DataObject(iofile=self.iofile)
        # Connecting cycle print to update_cycle func
        self.dataobject.cycle_printed.connect(self.update_cycle)
        self.dataobject.return_coords.connect(self.update_fig)
        self.send_settings.connect(self.dataobject.get_xyz_data)
        self.start_cmd.connect(self.dataobject.walk)
        self.dataobject.moveToThread(self.thread)
        # Starting thread
        self.thread.start()
        # Starting QObject's walk_Dirs in thread
        self.start_cmd.emit('')

    def _load_data(self, pfile):
        """Load saved data.

        Function that loads simulation data from a pre-compiled
        file and initializes the VisualApp window for plotting said data.
        """
        if pfile.endswith(('.hdf5', '.zip')):
            self.statusbar.showMessage('Loading data from compressed file.')
        else:
            self.statusbar.showMessage('Loading data from trajectory.')
        # Set-up data object using PathDensity class and moving to thread
        self.thread = QtCore.QThread()
        self.dataobject = VisualObject(pfile=pfile,
                                       trajfile=self.trajfile)
        # Connecting cycle print to update_cycle func
        self.dataobject.cycle_printed.connect(self.update_cycle)
        self.dataobject.return_coords.connect(self.update_fig)
        # Function takes hdf5 formats
        self.start_cmd.connect(self.dataobject.get_data)
        self.send_settings.connect(self.dataobject.get_xyz_data)
        self.dataobject.moveToThread(self.thread)
        # Starting thread
        self.thread.start()
        self.start_cmd.emit(pfile)

    def _reload(self):
        """Reload the data of VisualApp.

        Function that clears the old data of VisualApp and initializes
        the load of new.
        """
        self.folder, self.iofile = None, None
        self.trajfile, self.rst_file = None, None
        self.dataobject = None
        self.toggle_buttons(False)
        self._load_file()

    def _init_widget(self):
        """Initialize QDropWidget.

        Function initializing the settings of the QDropWidget of VisualApp
        window. Adds correct items to the ComboBoxes of op1,op2 and E, and
        add the CustomFigCanvas object to the QFrame.
        If no folder or file is given when initializing VisualApp, a
        QFileDialog is opened to get file/folder.
        """
        self.statusbar.showMessage('No data loaded')
        # Actions and menubar
        self.actionExit.triggered.connect(self.close_event)
        # Save figure as ...
        self.action_png.triggered.connect(self.save_png)
        self.action_makefig_script.triggered.connect(self.save_script)
        # Save figure data as ...
        self.action_hdf5.triggered.connect(self.save_hdf5)
        self.action_json.triggered.connect(self.save_json)
        self.action_datafile.triggered.connect(self.save_textdata)
        self.action_Load_data.triggered.connect(self.action_reload)
        # Save data as ...
        self.action_sim_hdf5.triggered.connect(self._save_sim_data_hdf5)
        self.action_sim_hdf5_zip.triggered.connect(self._save_sim_data_hdf5_z)
        # Connect show reg.line and interfaces, and method check
        self.intShowChkBtn.stateChanged.connect(self.toggle_intf)
        self.regLineChkBtn.stateChanged.connect(self.toggle_regl)
        self.cbarChkBtn.stateChanged.connect(self.toggle_cbar)
        self.showTitleChkBtn.stateChanged.connect(self.toggle_titles)
        self.plotTypeComBox.activated.connect(self.disable_zcombox)
        self.plotTypeComBox.activated.connect(self.disable_reglinecheck)
        # The Frame
        self.mainFrame.setStyleSheet(
            "QWidget { background-color: %s }" % QtGui.QColor(
                210, 210, 235, 255).name()
        )
        self.layout_a = QtWidgets.QGridLayout()
        self.mainFrame.setLayout(self.layout_a)
        # The Figure
        self.myfig = CustomFigCanvas()
        self.layout_a.addWidget(self.myfig)
        # Setting up the dropwidget of VisualApp
        self.show()
        # Refresh button
        self.refreshBtn.clicked.connect(self.refresh_data)
        # Bind update button to function
        self.updateBtn.clicked.connect(self.emit_settings)
        self.saveFigBtn.clicked.connect(self.save_png)
        self.previewBtn.clicked.connect(self.update_preview)
        # Bind <ENTER>-press in colormap line to function
        self.cmapComBox.lineEdit().returnPressed.connect(self._change_cmap)
        # Bind changes to title and labels to functions
        self.titleSizeSpin.valueChanged.connect(self._update_canvas_font)
        self.axesSizeSpin.valueChanged.connect(self._update_canvas_font)
        self.titleLine.returnPressed.connect(self._update_canvas_text)
        self.xaxisLabel.returnPressed.connect(self._update_canvas_text)
        self.yaxisLabel.returnPressed.connect(self._update_canvas_text)
        self.zaxisLabel.returnPressed.connect(self._update_canvas_text)
        # Show only stored trjs
        self.storedTrjsChkBtn.stateChanged.connect(
            self.toggle_only_stored_trjs)
        # Interactiv tab
        self.previewBtn_2.clicked.connect(self.emit_settings)
        self.drawButton.clicked.connect(self.draw_trajectory)
        self.playBtn.clicked.connect(self.play_cycle)
        self.store_trj_btn.clicked.connect(self.create_full_traj)
        # Recalculation
        self.actionRecalculate_data.triggered.connect(self.recalculate)
        # Analysis tab
        self.clusterButton.clicked.connect(self.statistic_analysis)
        self.correlationBtn.clicked.connect(self.show_correlation)
        self.RandForestBtn.clicked.connect(self.random_forest)
        self.DecisionTreeBtn.clicked.connect(self.decision_tree)
        self.PCABtn.clicked.connect(self.initiate_pca)

    def _change_cmap(self):
        """Set colormap of surface.

        Function that tries to set the colormap of object myfig.surf (plot)
        Bound to detection of returnPress in colormap combo box.
        """
        meth = self.settings['method']
        col = self.cmapComBox.currentText()

        try:
            # Test if entered colormap is valid
            _ = plt.get_cmap(col)
            if meth[0] == 'Density':
                # Method is 2dhist-plot/list where 4th item is the plot/surf.
                self.myfig.surf[3].set_cmap(col)
            elif meth[0] == 'Scatter':
                # If scatter, replot.
                self.statusbar.showMessage(
                    'Scatter plots have to be re-drawn to update colors!')
            else:
                # Should work for most surf-objects.
                self.myfig.surf.set_cmap(col)
        except AttributeError:
            # If surf has no .set_cmap()
            self.statusbar.showMessage(
                'Could not recognize colormap, try to "Update" figure...')
        except TypeError:
            # If no surf-object, or NoneType in general
            self.statusbar.showMessage(
                'Figure not recognized, try to "Update" figure...')
        except ValueError:
            # Did not recognize 'col' in cm
            self.statusbar.showMessage(
                'Chosen colormap not recognized!')
        self.myfig.fig.canvas.draw()

    def _get_op_range(self):
        """Display the range of the order parameter to the GUI."""
        interfaces = self.dataobject.infos['interfaces']
        self.minDataRange.setText(str(interfaces[0]))
        self.maxDataRange.setText(str(interfaces[-1]))

    def _change_zoom(self):
        """Set the zoom/x-&y-limits of the plot."""
        self.statusbar.showMessage('Drawing figure...')
        if self.settings['dim'] == 3:
            self.myfig.ax.set_xlim3d(*self.settings['x-limits'], )
            self.myfig.ax.set_ylim3d(*self.settings['y-limits'], )
            self.myfig.ax.set_zlim3d(*self.settings['z-limits'], )
        elif self.settings['dim'] == 2:
            self.myfig.ax.set_xlim(*self.settings['x-limits'], )
            self.myfig.ax.set_ylim(*self.settings['y-limits'], )
        # Change range on colorbar
        # Colorbar of contour plots does not scale, but color is updated
        norm = mpl.colors.Normalize(vmin=self.settings['z-limits'][0],
                                    vmax=self.settings['z-limits'][1])
        self.myfig.cbar.mappable.set_norm(norm=norm)
        self.myfig.cbar.draw_all()
        self.myfig.fig.canvas.draw()
        self.statusbar.showMessage('Plot ready!')

    def update_preview(self):
        """Update the visual of the plot.

        Button function, updates visual of plot based on all settings of
        'Plot'-tab in VisualApp.
        """
        self.statusbar.showMessage('Drawing figure...')
        self._get_settings()
        self._update_canvas_text()
        self._change_zoom()
        self._change_cmap()
        self.myfig.fig.canvas.draw()
        self.statusbar.showMessage('Plot ready!')

    # Defining button functions that toggle the visibility of
    # interfaces, regression line, colorbar and title/axis labels.
    def toggle_intf(self):
        """Toggle interface visibility."""
        select = self.intShowChkBtn.isChecked()
        for line in self.myfig.intf:
            line.set_visible(select)
        self.myfig.fig.canvas.draw()

    def toggle_regl(self):
        """Toggle regression line and legend visibility."""
        if self.settings['dim'] == 2:
            select = self.regLineChkBtn.isChecked()
            self.myfig.regl[0].set_visible(select)
        self.myfig.fig.canvas.draw()

    def toggle_cbar(self):
        """Toggle colorbar visibility."""
        select = self.cbarChkBtn.isChecked()
        self.myfig.cbar_ax.set_visible(select)
        self.myfig.fig.subplots_adjust(left=0.1, right=0.85,
                                       bottom=0.1, top=0.9)
        self.myfig.fig.canvas.draw()

    def toggle_titles(self):
        """Toggle labels and titles visibility."""
        select = self.showTitleChkBtn.isChecked()
        self.myfig.title.set_visible(select)
        self.myfig.xaxis.set_visible(select)
        self.myfig.yaxis.set_visible(select)
        self.myfig.zaxis.set_visible(select)
        self.myfig.fig.canvas.draw()

    def toggle_only_stored_trjs(self):
        """Toggle to show only data with trajectory files."""
        self.emit_settings()

    def emit_settings(self):
        """Update settings before sending to the data object.

        Function calls for an update of data/plot settings before sending
        dictionary to data object in main worker to update the data lists
        x, y, (z).

        """
        self.toggle_buttons(False)
        # Updating status bar of VisualApp window
        self.statusbar.showMessage('Loading data...')
        self._get_settings()
        if 'Contour' in self.settings['method'] and \
                self.settings['E'] == 'None':
            self.statusbar.showMessage('Invalid combination, '
                                       'contours need z-values')
            self.toggle_buttons(True)
        elif self.settings['method'][0] in ['Surface', 'Contour'] and \
                self.settings['op1'] == self.settings['op2']:
            self.statusbar.showMessage('Invalid combination, '
                                       'two axes are identical')
            self.toggle_buttons(True)
        else:
            self.send_settings.emit(self.settings)

    @QtCore.pyqtSlot(list)
    def update_cycle(self, cycles):
        """Set upper and lower bound of cycles for data loading.

        Function bound to pyqtSignal from DataObject in QThread, sets upper
        and lower bound of cycles for data loading, and enables 'Update'
        QPushButton when finished.

        Parameters
        ----------
        cycles : list
            Integers defining the initial en final cycle number to plot.

        """
        # Adding items to ComboBoxes in VisualApp.Window
        self.firstOpComBox.clear()
        self.firstOpComBox.addItems(self.dataobject.infos['op_labels'])
        self.firstOpComBox.insertSeparator(
            len(self.dataobject.infos['op_labels']) + 1)
        self.firstOpComBox.addItems(ENERGYLABELS)
        self.secondOpComBox.clear()
        self.secondOpComBox.addItems(self.dataobject.infos['op_labels'])
        self.secondOpComBox.insertSeparator(
            len(self.dataobject.infos['op_labels']) + 1)
        labels = []
        for label in ENERGYLABELS:
            if label not in self.dataobject.infos['op_labels']:
                labels.append(label)
        self.secondOpComBox.addItems(labels)
        self.energyComBox.clear()
        self.energyComBox.addItems(self.dataobject.infos['op_labels'])
        self.energyComBox.insertSeparator(
            len(self.dataobject.infos['op_labels']) + 1)
        self.energyComBox.addItems(ENERGYLABELS)
        self.energyComBox.addItems(['None'])
        self.folderComBox.clear()
        self.folderComBox.addItems(self.dataobject.infos['ensemble_names'])
        self.folderComBox.insertSeparator(
            len(self.dataobject.infos['ensemble_names']) + 1)

        self.folderComBox.addItem('All')
        # Setting cycles in dropwidget
        minc, maxc = cycles[0], cycles[1]
        self.minCyclSpin.setRange(minc, maxc)
        self.maxCyclSpin.setRange(minc, maxc)
        self.minCyclSpin.setValue(minc)
        self.maxCyclSpin.setValue(maxc)
        # Enables the buttons and updates status
        self.statusbar.showMessage(
            'All %s cycles loaded from files, ready to plot!', maxc - minc)
        self.updateBtn.setEnabled(True)
        self.refreshBtn.setEnabled(True)
        self._get_op_range()

    @QtCore.pyqtSlot(list, list, list)  # noqa: C901
    def update_fig(self, x, y, z):
        """Update figure by pyqtSignal.

        Function that updates figure canvas of CustomFigCanvas and the
        colorbar. Initiated by pyqtSignal from running QThread.
        Function checks method chosen in GUI and calls on built-in functions
        of orderparam_density module to generate a surface/contour/scatter/etc.
        If checked, function also plots interfaces and regression line.

        Parameters
        ----------
        x, y, z : list
            Floats, the coordinates used in plotting.

        Updates/Draws
        -------------
        myfig.fig : Re-draws canvas with method, data, and possible
            interfaces, reg.line and legend.
        myfig.ax : As above. (Axes containing plot, legend, lines, etc)
        myfig.cbar_ax : Axes with colorbar of plot.

        """
        self.statusbar.showMessage('re-Drawing figure...')

        remove_nan([x, y, z])
        order_parameter = self.dataobject.infos['op_labels'][0]
        l_x, l_y, l_z = len(x), len(y), len(z)
        fail = False
        if l_x == 0 or l_y == 0:
            if self.storedTrjsChkBtn.isChecked():
                self.display_message_box('Cannot plot...',
                                         'No stored trajectory files.')
                self.storedTrjsChkBtn.setChecked(False)
            else:
                self.statusbar.showMessage(
                    'One (or more) lists are empty, plotting halted!' +
                    f'x: {l_x}, y: {l_y}')
            fail = True
        if 'Density' not in self.settings['method'][0]:
            if l_x != l_z or l_y != l_z:
                self.statusbar.showMessage(
                    'Lists have different lengths! ' +
                    f'x: {l_x}, y: {l_y}, z: {l_z}')
                fail = True
        if self.settings['dim'] == 3 and l_z == 0:
            self.statusbar.showMessage('Z list empty, nothing to plot!')
            fail = True

        if fail:
            self.updateBtn.setEnabled(True)
            self.refreshBtn.setEnabled(True)
            return

        # Do a myfig.set_up(), in case dim/projection has changed
        self.myfig.set_up(dim=self.settings['dim'],
                          cbar=self.settings['colorbar'])
        # Setting default xlim and ylim
        self._set_def_limits(x, y, z)

        plot = self.settings['method'][0].lower()
        if self.settings['method'][-1] == '(filled)':
            plot = plot + 'f'

        if self.settings['dim'] == 3:
            self.myfig.surf, self.myfig.cbar = gen_surface(
                x, y, z,
                self.myfig.fig,
                self.myfig.ax,
                cbar_ax=self.myfig.cbar_ax,
                dim=self.settings['dim'],
                method=plot,
                res_x=self.settings['res'],
                res_y=self.settings['res'],
                colormap=self.settings['colormap'])
        elif self.settings['dim'] == 2:
            if self.settings['method'][0] == 'Density':
                if l_x != l_y:
                    min_length = min(l_x, l_y)
                    x, y = x[:min_length], y[:min_length]
                    self.statusbar.showMessage(
                        'Lists have different lengths! Plotting smallest.. ' +
                        f'x: {l_x}, y: {l_y}')

                self.myfig.surf = self.myfig.ax.hist2d(
                    x, y, (self.settings['res'], self.settings['res']),
                    cmap=self.settings['colormap'], density=True)
                self.myfig.cbar = self.myfig.fig.colorbar(
                    self.myfig.surf[3], cax=self.myfig.cbar_ax)
                # Set zmin and zmax equal to min and max value in histogram
                self.zminLine.setText(str(self.myfig.surf[0].min()))
                self.zmaxLine.setText(str(self.myfig.surf[0].max()))
            else:
                self.myfig.surf, self.myfig.cbar = gen_surface(
                    x, y, z,
                    self.myfig.fig,
                    self.myfig.ax,
                    cbar_ax=self.myfig.cbar_ax,
                    dim=self.settings['dim'],
                    method=plot,
                    res_x=self.settings['res'],
                    res_y=self.settings['res'],
                    colormap=self.settings['colormap'])
        # Showing interface planes(3D)/lines(2D)
        self.myfig.intf = []
        if self.settings['op1'] == order_parameter:
            for i in self.dataobject.infos['interfaces']:
                if i > self.xmax:
                    break
                if self.settings['dim'] == 3:
                    interf = plot_int_plane(
                        self.myfig.ax,
                        i, (self.ymin, self.ymax),
                        (self.zmin, self.zmax),
                        visible=self.settings['show_int'])
                else:
                    interf = self.myfig.ax.axvline(
                        i, linewidth=1,
                        color='grey', alpha=0.5,
                        visible=self.settings['show_int'])

                self.myfig.intf.append(interf)

        # Showing regression line(2D only)
        self.myfig.regl = []
        if self.settings['dim'] == 2:
            self.myfig.regl = plot_regline(self.myfig.ax, x, y)
            self.myfig.regl[0].set_visible(self.settings['reg_line'])
            self.myfig.legend = self.myfig.ax.legend()
            self.myfig.legend.set_visible(self.settings['reg_line'])
            self.myfig.cbar_ax.set_visible(self.settings['colorbar'])
        self._get_titles()
        # Update CustomFigure canvas and status
        self.myfig.fig.canvas.draw()
        self.statusbar.showMessage('Plot ready!')
        self.toggle_buttons(True)

    def on_pick(self, event):
        """Write all information about the closest point to slots in PyVisA.

        Parameters
        ----------
        event : Matplotlib mouse event
            It activates the interactive function on the plot.

        """
        # If no data has been loaded
        if not self.dataobject:
            self.display_message_box('Cannot select!',
                                     'No data has been loaded.')
            return

        # Avoid picking with 3D plots
        self._get_settings()
        if self.settings['dim'] == 3:
            return

        # Remove old points
        if self.myfig.ax.lines:
            self.myfig.ax.lines.pop()

        # Avoid MemoryError from scipy cdist
        try:
            closest, cycle, ensemble = self.dataobject. \
                find_cycle_ensemble(event)
        except MemoryError:
            self.display_message_box("Error...",
                                     "Dataset is too large")
            return

        # If no data is plotted
        if len(closest) == 0:
            self.display_message_box("Error...",
                                     "Please plot data before picking")
            return

        self.xLine.setText(str(round(closest[0], 2)))
        self.yLine.setText(str(round(closest[1], 2)))
        self.cycLine.setText(str(cycle))
        self.ensLine.setText(ensemble)
        traj = self.dataobject.traj_data[ensemble][cycle]
        if 'status' in traj.info:
            self.statusLine.setText(traj.info['status'])
        if 'MC-move' in traj.info:
            self.mcLine.setText(traj.info['MC-move'])

        self.myfig.ax.plot(closest[0], closest[1],
                           color=self.colorBox.currentText(),
                           marker='o',
                           markersize=self.sizeSpinBox.value(),
                           label='point')
        self.myfig.fig.canvas.draw()

    def draw_trajectory(self):
        """Draw chosen trajectory on plot.

        Show a full chosen trajectory on the plot in a selected
        color with choices of marker size and line thickness.

        """
        colors = {'Green': 'g', 'Blue': 'b', 'Red': 'r',
                  'Cyan': 'c', 'Magenta': 'm', 'Yellow': 'y',
                  'Black': 'k', 'White': 'w'}

        # Remove old figures in the plot
        if self.myfig.ax.lines:
            self.myfig.ax.lines.pop()

        if not self.xLine.text():
            self.display_message_box('Cannot draw...', 'Select a trajectory')
            return

        self.myfig.fig.canvas.draw()
        criteria = {'cycles': [int(self.cycLine.text()),
                               int(self.cycLine.text())],
                    'ensemble_name': self.ensLine.text(),
                    'x': self.settings['op1'],
                    'y': self.settings['op2'],
                    'status': self.settings['ACC']}
        x, y, _, _ = self.dataobject.load_traj(criteria)
        if x and y:
            self.myfig.ax.plot(x, y,
                               color=colors[self.colorBox.currentText()],
                               marker='o',
                               linestyle='-',
                               linewidth=self.widthSpinBox.value(),
                               markersize=self.sizeSpinBox.value())
            self.myfig.fig.canvas.draw()
            self.statusbar.showMessage('Plot ready!')
        else:
            self.statusbar.showMessage(
                'Different length of lists, drawing halted!' +
                ' x: %s, y: %s', self.settings['op1'], self.settings['op2'])

    def create_full_traj(self, cycle=None, ensemble_name=None, rst_file=None):
        """Create trajectory file.

        Parameters
        ----------
        cycle: string, optional
            Cycle number of trajectory.
        ensemble_name: string, optional
            Ensemble name of trajectory.
        rst_file: string, optional
            Path to the input/output file of the simulation.

        """
        # If the function has been called through the save_trj button
        if not (cycle or ensemble_name or rst_file):
            cycle = self.cycLine.text()
            ensemble_name = self.ensLine.text()
            if not cycle:
                self.display_message_box('Cannot store trajectory!',
                                         'Please select a trajectory first.')
                return None
            if not isinstance(self.iofile, PathDensity) and \
                    self.iofile.endswith('.rst'):
                rst_file = self.iofile
            elif self.rst_file:
                rst_file = self.rst_file
            else:
                self.display_message_box('Cannot store trajectory!',
                                         'Please load PyVisA with a .rst '
                                         'file in the simulation directory.')
                return None

        # Find trajectory in simulation directory
        traj = self.dataobject.traj_data[ensemble_name][int(cycle)]
        if 'status' not in traj.info.keys() or not traj.info['stored']:
            self.display_message_box('Trajectory files not present.',
                                     'Pick another trajectory')
            return None

        # Find path to folders
        trr_status = 'traj-acc' if traj.info['status'] == 'ACC'\
            else 'traj-rej'

        path_to_cycle = os.getcwd() +\
            f'/{ensemble_name}/traj/{trr_status}/{cycle}'
        path_to_traj = path_to_cycle + '/traj'
        # Find the sequence and velocity of files in the trajectory
        files = read_traj_txt_file(path_to_cycle + '/traj.txt')

        # Find the topology file from rst_file
        trj_list = []
        input_settings = \
            settings.parse_settings_file(rst_file)['engine']['input_files']
        if input_settings['conf'].endswith('.g96'):
            self.display_message_box('Cannot load file..',
                                     'Configuration format not supported.')
            return input_settings['conf']

        # Create the correct order of traj-files
        for _, value in files.items():
            if not value[0].endswith(tuple(TRJ_FORMATS)):
                continue
            if int(value[1]) == -1:
                trj_list.append(self.rev_trj(md.load(
                    path_to_traj + f'/{value[0]}', top=input_settings['conf']
                )))
            else:
                trj_list.append(md.load(
                    path_to_traj + f'/{value[0]}', top=input_settings['conf']
                ))

        trj = trj_list[0]
        for _ in trj_list[1:]:
            trj += _

        trj_name = f'trajectory_{ensemble_name}_{cycle}.pdb'
        trj.save_pdb(trj_name)
        return trj_name

    def play_cycle(self):
        """Collect traj-files in order to animate the trajectory."""
        # Find cycle and ensemble name of chosen trj
        cycle = self.cycLine.text()
        ensemble_name = self.ensLine.text()

        if not cycle or not ensemble_name:
            self.display_message_box('Cannot play...',
                                     'No trajectory chosen.')
            return

        if not isinstance(self.iofile, PathDensity) and \
                self.iofile.endswith('.rst'):
            trj_name = self.create_full_traj(cycle, ensemble_name,
                                             self.iofile)
        elif self.rst_file:
            trj_name = self.create_full_traj(cycle, ensemble_name,
                                             self.rst_file)
        else:
            self.display_message_box('Invalid choice..',
                                     'Please load an rst file')
            return

        if trj_name:
            try:
                subprocess.run(["avogadro2", trj_name], check=True)
                os.remove(os.getcwd() + '/' + trj_name)
            except FileNotFoundError:
                self.display_message_box('Cannot play cycle',
                                         'Could not start avogadro2.' + '\n' +
                                         ' Make sure avogadro is installed' +
                                         ' and try again.')
                return

    @staticmethod
    def rev_trj(trj):
        """Reverse a trj via MD traj.

        Parameters
        ----------
        trj : mdtraj object
            Mdtraj MD trajectory.

        Returns
        -------
        r_trj : mdtraj object
            The reversed MD trajectory.

        """
        f_n = np.arange(len(trj) - 1, 0, -1)
        r_trj = trj.slice(f_n)
        return r_trj

    def display_message_box(self, line1, line2):
        """Display a message box with chosen text.

        Parameters
        ----------
        line1: string
            Title of message box.
        line2: string
            Main message for the box.

        """
        msg = QtWidgets.QMessageBox.question(self, line1,
                                             line2,
                                             QtWidgets.QMessageBox.Ok)
        if msg == QtWidgets.QMessageBox.Ok:
            try:
                return
            except AttributeError:
                pass

    def recalculate(self):
        """Recalculate OP and new CV's.

        Using the functions from recalculate_order.py this function
        will calculate new collective variables to the simulation while
        keeping the original OP in order to explore new descriptors of the
        system. The new files will replace the order files so PyVisA can
        load and visualize the new descriptors. The old files will
        be renamed as order_old.

        """
        msg = QtWidgets.QMessageBox.question(
            self, "Warning: Confirm descriptor recalculation...",
            "This operation might result in data-loss, " +
            "\n" + "are you sure to continue?",
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

        if msg == QtWidgets.QMessageBox.Yes:

            if not self.rst_file or 'rst' not in self.rst_file:
                self.display_message_box('Cannot perform re-calculations...',
                                         'Please load PyVisA with a rst file.')
                return
            if not recalculate_all(self.folder, self.rst_file,
                                   self.dataobject.infos['ensemble_names']):
                self.display_message_box('Cannot perform re-calculations...',
                                         'Invalid input.')
                return
            self.statusbar.showMessage('Recalculating...')
            self.statusbar.showMessage('Plot ready!')
            self.refresh_data()
            self.display_message_box('Recalculation complete!',
                                     'The new data can now be visualized.')

    def statistic_analysis(self):
        """Perform statistical analysis on plotted data.

        Function which performs a range of different clustering
        techniques like K-mean, Hierarchical, Gaussian Mixture,
        Spectral and PCA by choice of the user. The user specifies
        the type of analysis as well as the amount of cluster/components.

        """
        self.statusbar.showMessage('Performing analysis...')
        plt.ion()
        if not self.dataobject:
            self.display_message_box('Cannot analyse!',
                                     'No data has been loaded.')
            return

        self._get_settings()
        n_clusters = self.nrOfCluster.value()
        method = self.clusterTypeComBox.currentText()
        cmap = self.cmapComBox.currentText()

        if self.settings['dim'] == 3:
            self.display_message_box('Cannot perform statistics!',
                                     'Please choose a 2D plot.')
            return
        # If no data has been plotted:
        if not self.dataobject.settings:
            self.display_message_box('Cannot perform analysis...',
                                     'No data has been selected.')
            return
        data = np.column_stack([self.dataobject.x,
                                self.dataobject.y])

        methods = {'Kmeans': k_means,
                   'Hierarchical': hierarchical,
                   'Gaussian Mixture': gaussian_mixture,
                   'Spectral': spectral}
        try:
            methods[method](n_clusters, data, self.settings, cmap)
        except MemoryError:
            self.display_message_box('Dataset too large',
                                     'Cannot perform fitting.')
            return
        plt.ioff()
        self.statusbar.showMessage('Analysis completed!')

    def initiate_pca(self):
        """Perform PCA."""
        self._get_settings()
        n_pca = self.nrOfpca.value()
        cmap = self.cmapComBox.currentText()
        if n_pca > len(self.dataobject.infos['op_labels']) + 2:
            self.display_message_box('Cannot perform PCA, ',
                                     'The amount of components exceeds' +
                                     '\n' + 'the amount of features in'
                                            'the dataset')
        else:
            # Collect data for analysis
            data = self.dataobject.load_to_df(self.settings)
            try:
                pyvisa_pca(n_pca, self.settings, data, cmap)
            except ValueError:
                self.display_message_box('Cannot perform analysis!',
                                         'No data could be used with the '
                                         'chosen criteria. Try to change '
                                         'the status type of MC-move or '
                                         'type of path.')

    def random_forest(self):
        """Create a random forest from simulation results."""
        self.statusbar.showMessage('Performing analysis...')
        self._get_settings()
        min_op = float(self.minDataRange.text())
        max_op = float(self.maxDataRange.text())
        try:
            data, reactive = \
                self.dataobject.remove_values(self.settings, min_op, max_op)
        except AssertionError:
            self.display_message_box('Cannot perform analysis!',
                                     'No data could be used with the chosen'
                                     ' criteria. Try to change the status, '
                                     'type of MC-move or type of path.')
            return
        if len(data) == 0:
            self.display_message_box('Cannot perform analysis!',
                                     'No data was valid within the '
                                     'chosen data range.')
            return

        depth = self.LayerSpinBox.value()
        random_forest(data, reactive, depth=depth)
        self.statusbar.showMessage('Analysis completed!')

    def decision_tree(self):
        """Create a decision tree from simulation results."""
        self.statusbar.showMessage('Performing analysis...')
        self._get_settings()
        min_op = float(self.minDataRange.text())
        max_op = float(self.maxDataRange.text())
        try:
            data, reactive = \
                self.dataobject.remove_values(self.settings, min_op, max_op)
        except AssertionError:
            self.display_message_box('Cannot perform analysis!',
                                     'No data could be used with the chosen'
                                     ' criteria. Try to change the status, '
                                     'type of MC-move or type of path.')
            return

        if len(data) == 0:
            self.display_message_box('Cannot perform analysis!',
                                     'No data was valid within the '
                                     'chosen data range.')
            return
        depth = self.LayerSpinBox.value()
        decision_tree(data, reactive, depth=depth)
        self.statusbar.showMessage('Analysis completed!')

    def show_correlation(self):
        """Show the correlation between simulation data.

        Function which plots a correlation matrix from the simulation
        data.

        """
        self.statusbar.showMessage('Performing analysis...')
        plt.ion()
        self._get_settings()
        if not self.dataobject:
            self.display_message_box('Cannot analyse!',
                                     'No data has been loaded.')
            return
        data = self.dataobject.load_to_df(self.settings)
        correlation_matrix(data)
        plt.ioff()
        self.statusbar.showMessage('Analysis completed!')

    def refresh_data(self):
        """Signal a walk of the data."""
        if not self.rst_file and \
                self.iofile.endswith(('.hdf5', '.zip', '.rst')):
            self.display_message_box('Cannot refresh.',
                                     'Please load PyVisA with a .rst file \n')
            return
        if self.trajfile and self.iofile.endswith('.rst'):
            self.dataobject = DataObject(iofile=self.iofile)
        elif self.rst_file:
            self.dataobject = DataObject(iofile=self.rst_file)
        else:
            self.dataobject = DataObject(iofile=self.iofile)
        self.dataobject.cycle_printed.connect(self.update_cycle)
        self.dataobject.return_coords.connect(self.update_fig)
        self.send_settings.connect(self.dataobject.get_xyz_data)
        self.dataobject.walk()


class CustomFigCanvas(FigureCanvas):
    """
    Class definition of the custom figure canvas used in VisualApp.

    Attributes
    ----------
    surf : Placeholder for object shown in main subplot of figure.
    ax : Main subplot (Axes object) of CustomFigCanvas.
    cbar_ax : Subplot (Axes object) of colorbar in CustomFigCanvas.
    intf : Placeholder for a list of interface lines/planes in the figure.
    regl : Placeholder for 2D regression line in figure.
    legend : Placeholder for legend in figure.
    title : Placeholder for the figure title.
    {x/y/z} axis : Placeholder for the x/y/z-axis labels.

    """

    def __init__(self, ):
        """Initialize the FigureCanvas class."""
        self.fig = Figure()
        FigureCanvas.__init__(self, self.fig)
        self.surf = None
        self.cbar = None
        self.cbar_ax = None
        self.intf = None
        self.regl = None
        self.legend = None
        self.title = None
        self.ax = None
        self.xaxis, self.yaxis, self.zaxis = None, None, None
        self.set_up()

    def set_up(self, dim=2, cbar=True):
        """Define subplot projection (2D/3D).

        Function that defines the main subplot's projection (2D/3D)
        depending on chosen method from VisualApp GUI.

        Parameters
        ----------
        dim : integer, optional
              Dimension/projection of main subplot.
        cbar : boolean, optional
              If True, the space for the colorbar is considered
              in the setting.

        """
        # The Figure
        self.fig.clf()
        # Subplots, depending on dimension/projection
        if dim == 3:
            self.ax = self.fig.add_subplot(111, projection='3d')
        elif dim == 2:
            self.ax = self.fig.add_subplot(111)
        # Axes for colorbar
        self.cbar_ax = self.fig.add_axes([0.86, 0.1, 0.03, 0.8])
        if cbar:
            self.fig.subplots_adjust(left=0.1, right=0.85, bottom=0.1, top=0.9)
        else:
            self.fig.subplots_adjust(left=0.1, right=0.90, bottom=0.1, top=0.9)
        self.cbar_ax.set_visible(False)


class DataSlave(QtCore.QObject, PathVisualize):
    """QObject class definition that holds the PathDensity data.

    By pyqtSignals this object can be called upon to either do a directory
    "walk" to collect data, or get specific data into lists by specific
    settings.

    Attributes
    ----------
    Contains the same attributes as the PathDensity() class of
    orderparam_density.py in source.

    Functions
    ---------
    load_all: function
        Load x,y and z data from all trajectories selected by the settings
        of PyVisA. Also stored the coordinates, ensemble name and cycle
        of each point for use in interactivity in the plot.
    find_cycle_ensemble : function
        Accept a MouseClick event on the
        plot and return the x, y coordinate, cycle and ensemble name of the
        closest point to wherever the user clicks on the plot.
    load_to_df : function
        Loads trajectory data into a pandas dataframe for analysis.

    Signals
    -------
    cycle_printed : PyQt Signal
        Send a list of min and max cycle number to
        the VisualApp to update the dropwidget with correct values.
    returns_coords : PyQt Signal
        Send three lists of coords (or empty)
        to the VisualApp to update the figure.

    Slots
    -----
    walk : PyQt Slot connected to VisualApp
        When activated with a string,
        initializes the walk_Dirs() function of DataObject/PathDensity
        class. At end of function, emits list of cycles to signal
        cycle_printed.
    get : PyQt Slot connected to VisualApp
        When activated with a dictionary
        with relevant plotting settings, preforms a data fetch of either
        get_Edata or get_Odata on DataObject/PathDensity class.
        At end of function, emits lists of coords to signal return_coords.

    """

    return_coords = QtCore.pyqtSignal(list, list, list)

    def __init__(self, iofile=None):
        """Initialize the QObject class and classes inherited.

        Parameters
        ----------
        iofile : string, optional
            The input file name.

        """
        self.iofile = iofile
        QtCore.QObject.__init__(self)
        PathVisualize.__init__(self)
        self.settings = {}
        self.x, self.y, self.z = [], [], []
        self.data_origin = []

    def load_all(self):
        """Load chosen data from all trajectories.

        Function that uses the load_traj function from orderparam_density()
        to load all the x, y, and z data corresponding to the chosen
        settings for plotting. The criteria for plotting is collected
        from the settings and sent to load_traj. It also stores the
        coordinates, and data_origin of each points which is used for
        the interactivity in other functions like on_pick.

        Returns
        -------
        x : list
            List of data from chosen parameter.
        y : list
            List of data from chosen parameter.
        z : list
            List of data from chosen parameter.

        """
        criteria = {'x': self.settings['op1'],
                    'y': self.settings['op2'],
                    'z': self.settings['E'],
                    'ensemble_name': self.settings['fol'],
                    'cycles': (self.settings['min_cycle'],
                               self.settings['max_cycle']),
                    'status': self.settings['ACC'],
                    'MC-move': self.settings['MC-move'],
                    'weight': self.settings['weight'],
                    'stored': self.settings['stored']}

        if self.settings['start_end'] == 'All':
            pass
        else:
            criteria['reactive'] = \
                self.settings['start_end'].lower() == 'reactive'

        x, y, z = [], [], []
        data_origin = []
        if criteria['ensemble_name'] == 'All':
            for fol in self.infos['ensemble_names']:
                criteria['ensemble_name'] = fol
                x_i, y_i, z_i, doi = self.load_traj(criteria)
                x.extend(x_i)
                y.extend(y_i)
                z.extend(z_i)
                data_origin.extend(doi)
        else:
            x, y, z, data_origin = self.load_traj(criteria)

        self.x, self.y, self.z = x, y, z
        self.data_origin = data_origin
        return x, y, z

    @QtCore.pyqtSlot(dict)
    def get_xyz_data(self, new_settings):
        """Signal emission to get xyz data from settings in VisualApp.

        Parameters
        ----------
        new_settings : dict
            The settings provided by an user to select the data to extract.

        """
        self.settings = new_settings
        x, y, z = self.load_all()

        # If allowing data shift
        if self.settings['try_shift']:
            op1 = self.settings['op1']
            fixedx = op1 == self.infos['op_labels'][0]
            x, y = try_data_shift(x, y, fixedx)
        self.return_coords.emit(x, y, z)

    def find_cycle_ensemble(self, event):
        """Return x, y, cycle and ensemble name at the closest point.

        Parameters
        ----------
        event : Matplotlib mouse event
            The users mouse click on the plot.

        Returns
        -------
        closest : list
            List of the x and y coordinate of the closest point
            to the users click on the plot.
        cycle : int
            Corresponding cycle of the closest point.
        ensemble_name : string
            Corresponding ensemble name of the closest point.

        """
        # If no data is plotted
        if not (hasattr(self, 'x') and hasattr(self, 'y')):
            return [], '', ''
        if len(self.x) == 0 and len(self.y) == 0:
            return [], '', ''

        # Find the closest point
        data_list = list(zip(self.x, self.y))
        data_list = np.array(data_list)
        closest_index = distance.cdist([[event.xdata, event.ydata]],
                                       data_list,
                                       'seuclidean').argmin()

        closest = data_list[closest_index]
        # Find corresponding cycle and ensemble
        ensemble_name = self.data_origin[closest_index][0]
        cycle = self.data_origin[closest_index][1]
        return closest, cycle, ensemble_name

    def load_to_df(self, sim_settings):
        """Create a pandas dataframe for analysis.

        Parameters
        ----------
        sim_settings : dict
            Dict containing settings from GUI.

        Returns
        -------
        df : Pandas Dataframe
            Dataframe containing all OP, CV and energy data from
            selected ensembled.

        """
        dataframe = pd.DataFrame()
        number_variables = len(self.infos['op_labels']) + 3

        selection_crit = {'stored': sim_settings['stored'],
                          'reactive': sim_settings['start_end'] == 'Reactive'}
        if sim_settings['ACC'] != 'BOTH':
            selection_crit['status'] = sim_settings['ACC']
        if sim_settings['MC-move'] != 'All':
            selection_crit['MC-move'] = sim_settings['MC-move']

        ensemble_names = self.traj_data.keys() if \
            sim_settings['fol'] == 'All' else [sim_settings['fol']]
        for ensemble_name in ensemble_names:
            for cycle in self.traj_data[ensemble_name].keys():
                traj = self.traj_data[ensemble_name][int(cycle)]
                # If columns are missing from the trajectory
                if len(traj.frames.columns) == number_variables:
                    for key in selection_crit:
                        if selection_crit.get(key, 1) != traj.info.get(key, 2):
                            break
                    else:
                        dataframe = dataframe.append(traj.frames,
                                                     ignore_index=True,
                                                     sort=False)
        dataframe = dataframe.dropna()
        return dataframe

    def load_reactive_unreactive(self, sim_settings):
        """Assign a boolean for reactive (True) and unreactive (False) path.

        Parameters
        ----------
        sim_settings : dict
            Dict containing settings from GUI.

        Returns
        -------
        reactive_data : Pandas Dataframe
            Dataframe containing True/False for each frame in the simulation.
            True if the frame is reactive, False if unreactive.

        """
        reactive_data = pd.DataFrame()
        number_variables = len(self.infos['op_labels']) + 3

        selection_crit = {'stored': sim_settings['stored'],
                          'reactive': sim_settings['start_end'] == 'Reactive'}
        if sim_settings['ACC'] != 'BOTH':
            selection_crit['status'] = sim_settings['ACC']
        if sim_settings['MC-move'] != 'All':
            selection_crit['MC-move'] = sim_settings['MC-move']

        if sim_settings['fol'] == 'All':
            ensemble_names = self.traj_data.keys()
        else:
            ensemble_names = [sim_settings['fol']]
        for ensemble_name in ensemble_names:
            for cycle in self.traj_data[ensemble_name].keys():
                traj = self.traj_data[ensemble_name][int(cycle)]
                if len(traj.frames.columns) == number_variables:
                    for key in selection_crit.items():
                        if selection_crit[key] != traj.info[key]:
                            break
                    else:
                        reactive_list = \
                            [traj.info['reactive']] * traj.info['length']
                        reactive_data = reactive_data.append(
                            reactive_list, ignore_index=True, sort=False)
        return reactive_data

    def remove_values(self, sim_settings, min_op, max_op):
        """Remove values below and above chosen values in a dataframe.

        Parameters
        ----------
        sim_settings : dict
            Dict containing the simulation settings.
        min_op : float
            Lowest chosen value of the order parameter.
        max_op : float
            highest chosen value of the order parameter.

        Returns
        ------
        sim_reduced : Dataframe
            Dataframe containing data from the simulation with removed frames
            where the op is above or below a predetermined value.
        reactive_reduced : Dataframe
             Dataframe containing the reactive/unreactive flag for each
            frame in the simulation with removed frames
            where the op is above or below a predetermined value.

        """
        data = self.load_to_df(sim_settings)
        reactive = pd.DataFrame(np.array(self.load_reactive_unreactive(
            sim_settings), dtype=bool), columns=['reactive'])
        op_label = self.infos['op_labels'][0]
        tot_data = data.join(reactive)
        tot_data = tot_data[~(tot_data[op_label] <= min_op)]
        tot_data = tot_data[~(tot_data[op_label] >= max_op)]
        tot_data = tot_data.reset_index(drop=True)
        reactive_reduced = tot_data['reactive']
        data_reduced = tot_data.drop('reactive', axis=1)
        return data_reduced, reactive_reduced


class VisualObject(DataSlave):
    """DataSlave class used by VisualApp to load data from compressed file."""

    cycle_printed = QtCore.pyqtSignal(list)

    def __init__(self, pfile=None, trajfile=None):
        """Initialize the class and classes inherited.

        Parameters
        ----------
        pfile : string, optional
            The input file name.
        trajfile : string, optional
            The input trajectory file name.

        """
        self.pfile = pfile
        self.trajfile = trajfile
        DataSlave.__init__(self)
        self.settings = {}

    def load_data_from_single(self, iofile):
        """Load a single trajectory file.

        Parameters
        ----------
        iofile : string
            The input file name.

        """
        # If the -data command has been used
        if self.trajfile:
            trajectory_file = self.trajfile
            search_dir = os.path.dirname(
                os.path.abspath(self.trajfile))
        else:
            # Trajectory file selected through the load data option
            trajectory_file = iofile
            search_dir = os.path.dirname(
                os.path.abspath(self.pfile))

        # Find ensemble name and cycle if possible
        ensemble_name, cycle = 'trj', 0
        for part in search_dir.split('/'):
            if part.startswith('0') and len(part) == 3:
                ensemble_name = part
            elif part.isnumeric():
                cycle = int(part)

        # Find the settings from rst files
        if iofile.endswith('.rst'):
            input_settings = settings.parse_settings_file(
                os.path.abspath(iofile))
        else:
            input_settings = settings.parse_settings_file(
                os.path.abspath(find_rst_file(search_dir)))

        # Create a composite order parameter
        functions = create_orderparameter(input_settings)
        frames = {}

        # Calculate collective variables and order parameter
        for idx, order_p in enumerate(
                recalculate_order(functions, trajectory_file, {})):
            frames[idx] = order_p

        # Create data frames
        cols = []
        for i in range(1, len(frames[0]) + 1):
            cols.append(f'op{i}')
        op_labels = get_cv_names(input_settings)
        labels = op_labels if len(op_labels) == len(cols) else cols
        frames = pd.DataFrame.from_dict(frames,
                                        orient='index',
                                        columns=labels)

        # Fill information to info and infos
        info = {'ordermax': np.max(frames[labels[0]]),
                'ordermin': np.min(frames[labels[0]]),
                'length': len(frames.index),
                'ensemble_name': ensemble_name,
                'cycle': cycle}
        # Create Trajectory with frames and info
        traj = Trajectory(frames, info)
        self.traj_data = {ensemble_name: {cycle: traj}}
        infos = {'ensemble_names': [ensemble_name],
                 'cycles': [cycle, cycle],
                 'op_labels': traj.frames.columns,
                 'interfaces': input_settings['simulation']['interfaces']}
        self.infos = infos

    @QtCore.pyqtSlot(str)
    def get_data(self, pfile):
        """Signal emission for loading data from file.

        Parameters
        ----------
        pfile : string
            The input file name.

        """
        self.pfile = pfile
        if self.pfile.endswith(tuple(TRJ_FORMATS)) or \
                (self.trajfile and self.trajfile.endswith(tuple(TRJ_FORMATS))):
            self.load_data_from_single(self.pfile)
            if not self.infos:
                return
        elif pfile.endswith(('.hdf5', '.zip')):
            self.load_whatever()
        else:
            raise ValueError('Format not recognized')
        cycles = self.infos['cycles']
        self.cycle_printed.emit(cycles)


class DataObject(DataSlave, PathDensity):
    """DataSlave class used by VisualApp to load data from PyRETIS input."""

    cycle_printed = QtCore.pyqtSignal(list)

    def __init__(self, iofile=None):
        """Initialize the class and classes inherited.

        Parameters
        ----------
        iofile : string, optional
            The input file name.

        """
        DataSlave.__init__(self, iofile=iofile)
        PathDensity.__init__(self, iofile=iofile)
        self.traj_data = self.traj_dict

    @QtCore.pyqtSlot(str)
    def walk(self):
        """Signal emission to begin 'walk' in the simulation directory."""
        if 'cycles' not in self.infos:
            if 'only_ops' in self.infos.keys():
                self.walk_dirs(self.infos['only_ops'])
            else:
                self.walk_dirs()
        cycles = self.infos['cycles']
        self.cycle_printed.emit(cycles)


def visualize_main(basepath, infile=None, rst_file=None, trajfile=None):
    """Run the VisualApp application.

    Parameters
    ----------
    basepath : string
        The execution folder where the input files are.
    infile : string, optional
        Path to the data that PyVisA is initialized with. Can be a
        PathDensity object, or a hdf5 file.
    rst_file : string, optional
        Path to the rst file if PyVisA has been loaded through the
        input/output rst files of the simulation.
    trajfile: string, optional
        Path to the trajectory file is PyVisA has been loaded with
        the -data command.

    """
    app = QtWidgets.QApplication(sys.argv)
    sub_files = {'rst': rst_file, 'traj': trajfile}
    window = VisualApp(folder=basepath, iofile=infile, sub_files=sub_files)
    window.show()
    sys.exit(app.exec_())
